/* IPSec ESP and AH support.
   Copyright (c) 1999, 2002 Pierre Beyssac and Geoffrey Keating.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-
 * Copyright (c) 1999 Pierre Beyssac
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <poll.h>
#include <signal.h>

#include <openssl/evp.h>
#include <openssl/hmac.h>

#ifdef USE_ETHERTAP
/* Use ethertap device under Linux */
struct ethtap_header {
    unsigned short padding;
    unsigned char dst[6];
    unsigned char src[6];
    unsigned short type;
};

struct ethtap_header ethtap;

#define MAX_LINKHEADER	(sizeof(struct ethtap_header))
#else
#define MAX_LINKHEADER	0
#endif

#define max(a,b)	((a)>(b)?(a):(b))

#define MAX_HEADER	max(64,MAX_LINKHEADER)
#define MAX_SECRET	max(EVP_MAX_KEY_LENGTH, EVP_MAX_MD_SIZE);
#define MAX_PACKET	4096
#define MAX_PEERS	16

#define UDP_PORT	2001

unsigned char buf[MAX_HEADER+MAX_PACKET];

struct sa_desc {
    struct sa_desc *next;

    struct sockaddr_in init;	/* initial and fallback remote address */
    struct sockaddr_in dest;	/* current remote address */
    struct sockaddr_in source;	/* local socket address we send packets from */
    unsigned char use_fallback;	/* use initial address as fallback? */
    unsigned char use_dest;	/* is dest address known yet? */

    unsigned long spi;		/* security parameters index */
    unsigned long seq_id;	/* for replay protection (not implemented) */

    /* Encryption key */
    unsigned char enc_secret[EVP_MAX_KEY_LENGTH];
    unsigned int enc_secret_size;
    /* Preprocessed encryption key */
    EVP_CIPHER_CTX enc_key;

    /* Authentication secret */
    unsigned char auth_secret[EVP_MAX_MD_SIZE];
    unsigned int auth_secret_size;
    /* Authentication method to use, or NULL */
    EVP_MD *hm;

    /* Encapsulation method to use to send packets */
    struct encap_method *em;

    /* flags */
    unsigned char local, no_iv;
    /* timeout counters */
    time_t last_packet_sent, last_packet_recv, last_checkifaddr;
};

struct tun_method {
    unsigned int link_header_size;
    unsigned char *link_header;
};

struct peer_desc {
    struct sa_desc *local_sa, *remote_sa;
    int tun_fd;		/* file descriptor for associated tunnel device */
    struct tun_method *tm;
};

/* Size of sent hash (doesn't have to be the full hash) */
#define HMAC_SIZE	12

/* UDP encap header. Currently tries to mimic IPSEC's AH header */
typedef struct udp_encap_header {
    unsigned long spi;			/* security parameters index */
    unsigned long seq_id;		/* sequence id (unimplemented) */
    unsigned char hmac[HMAC_SIZE];	/* HMAC */
} udp_encap_header_t;

/* ICMP "custom AH" encap header */
typedef struct icmp_encap_header {
  /* ICMP fields */
    unsigned char icmp_type;
    unsigned char icmp_code;
    unsigned short icmp_cksum;
  /* Our custom fields */
    unsigned long spi;			/* security parameters index */
    unsigned long seq_id;		/* sequence id (unimplemented) */
    unsigned char hmac[HMAC_SIZE];	/* HMAC */
} icmp_encap_header_t;

/* A real AH header (RFC 2402) */
typedef struct ah_encap_header {
    unsigned char next_header;
    unsigned char payload_len;
    unsigned short reserved;
    unsigned long spi;			/* security parameters index */
    unsigned long seq_id;		/* sequence id (unimplemented) */
    unsigned char hmac[HMAC_SIZE];	/* HMAC */
} ah_encap_header_t;

/* A real ESP header (RFC 2406) */
typedef struct esp_encap_header {
    unsigned long spi;			/* security parameters index */
    unsigned long seq_id;		/* sequence id (unimplemented) */
    /* variable-length payload data + padding */
    /* unsigned char next_header */
    /* optional auth data */
} esp_encap_header_t;

struct encap_method {
    int fd;			/* file descriptor for relevant socket */
    unsigned char *name;

    int fixed_header_size;

    /* Description of the packet being processed */
    unsigned char *buf;
    unsigned int bufsize, bufpayload, var_header_size;
    int buflen;
    struct sockaddr_in from; int fromlen;

    int (*recv)(struct encap_method *encap,
		unsigned char *buf, unsigned int bufsize,
		struct sockaddr_in *from);
    int (*hmac_compute)(struct encap_method *encap,
			unsigned char do_store,
			struct in_addr *src_ip, EVP_MD *hm,
			unsigned char *secret, unsigned short secret_size);
    struct peer_desc *(*peer_find)(struct encap_method *encap);
    void (*send_peer)(struct encap_method *encap,
		      struct peer_desc *peer,
		      unsigned char *buf, unsigned int bufsize);
    int (*recv_peer)(struct encap_method *encap,
		     struct in_addr *src_ip,
		     struct peer_desc *peer);
};

/* Forward decl */
int hmac_udp_compute(struct encap_method *encap,
		     unsigned char do_store,
		     struct in_addr *src_ip, EVP_MD *hm,
		     unsigned char *secret, unsigned short secret_size);
int hmac_rawip_compute(struct encap_method *encap,
		       unsigned char do_store,
		       struct in_addr *src_ip, EVP_MD *hm,
		       unsigned char *secret, unsigned short secret_size);
int hmac_icmp_compute(struct encap_method *encap,
		      unsigned char do_store,
		      struct in_addr *src_ip, EVP_MD *hm,
		      unsigned char *secret, unsigned short secret_size);
void encap_udp_send_peer(struct encap_method *encap,
			 struct peer_desc *peer,
			 unsigned char *buf, unsigned int bufsize);
void encap_ah_send_peer(struct encap_method *encap,
			struct peer_desc *peer,
			unsigned char *buf, unsigned int bufsize);
void encap_esp_send_peer(struct encap_method *encap,
			 struct peer_desc *peer,
			 unsigned char *buf, unsigned int bufsize);
void encap_icmp_send_peer(struct encap_method *encap,
			  struct peer_desc *peer,
			  unsigned char *buf, unsigned int bufsize);
struct peer_desc *peer_find(unsigned long spi, struct encap_method *encap);
struct sa_desc *find_local_sa(unsigned long spi, struct encap_method *encap);
struct sa_desc *find_remote_sa(unsigned long spi, struct encap_method *encap);
int encap_hmac_recv_peer(struct encap_method *encap,
			 struct in_addr *src_ip,
			 struct peer_desc *peer);
int encap_esp_recv_peer(struct encap_method *encap,
			struct in_addr *src_ip,
			struct peer_desc *peer);

/* Yuck! Global variables... */

struct peer_desc peers[MAX_PEERS];

struct tun_method tm_tun;
#ifdef USE_ETHERTAP
struct tun_method tm_tap;
#endif

unsigned short peer_num = 0;
unsigned short sa_num = 0;
unsigned short ip_id;

/* Descriptors for predefined encapsulation methods */
#define ENCAP_IPAH	0
#define ENCAP_UDP	1
#define ENCAP_IPESP	2
#define ENCAP_ICMP	3
#define ENCAP_MAX	4
struct encap_method encap_meth[ENCAP_MAX];

/* Security associations lists */
struct sa_desc *local_sa_list = NULL;
struct sa_desc *remote_sa_list = NULL;

/* Interval for keepalive probes */
int keepalive_recv = 600;
int keepalive_send = 200;

int check_ifaddr = 60;

unsigned char global_iv[EVP_MAX_IV_LENGTH] = { 1 };

#define encap_get_fd(e)	((e)->fd)
#define encap_recv(e,b,bs,f) \
	((e)->recv((e),(b),(bs),(f)))
#define encap_hmac_compute(e,s,hm,se,ss) \
	((e)->hmac_compute((e),1,(s),(hm),(se),(ss)))
#define encap_hmac_cmp(e,s,hm,se,ss) \
	((e)->hmac_compute((e),0,(s),(hm),(se),(ss)))
#define encap_peer_find(e) \
	((e)->peer_find((e)))
#define encap_send_peer(e,p,b,bs) \
	((e)->send_peer((e),(p),(b),(bs)))
#define encap_recv_peer(e,s,p) \
	((e)->recv_peer((e),(s),(p)))

/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 */
u_short
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

/*
 * Find an encap method by name
 */
struct encap_method *find_encap(unsigned char *name)
{
    int i;
    for (i = 0; i < ENCAP_MAX; i++)
        if (strcmp(encap_meth[i].name, name) == 0)
	    return encap_meth+i;
    return NULL;
}

/*
 * Decapsulate from a UDP packet
 */
int encap_udp_recv(struct encap_method *encap,
		   unsigned char *buf, unsigned int bufsize,
		   struct sockaddr_in *from)
{
    int r;

    encap->fromlen = sizeof(encap->from);

    r = recvfrom(encap->fd, buf, bufsize, 0,
		 (struct sockaddr *)&encap->from, &encap->fromlen);
    if (r == -1) {
	syslog(LOG_ERR, "recvfrom: %m");
	return -1;
    }
    if (r < encap->fixed_header_size) {
        syslog(LOG_ALERT, "packet too short from %s",
		inet_ntoa(encap->from.sin_addr));
	return -1;
    }
    encap->buf = buf;
    encap->bufsize = bufsize;
    encap->buflen = r;
    encap->bufpayload = 0;
    *from = encap->from;
    return r;
}

/*
 * Decapsulate from a raw IP packet
 */
int encap_rawip_recv(struct encap_method *encap,
		     unsigned char *buf, unsigned int bufsize,
		     struct sockaddr_in *from)
{
    int r;
    struct ip *p = (struct ip *)buf;

    encap->fromlen = sizeof(encap->from);

    r = recvfrom(encap->fd, buf, bufsize, 0,
		 (struct sockaddr *)&encap->from, &encap->fromlen);
    if (r == -1) {
	syslog(LOG_ERR, "recvfrom: %m");
	return -1;
    }
    if (r < (p->ip_hl << 2)+encap->fixed_header_size) {
        syslog(LOG_ALERT, "packet too short from %s",
		inet_ntoa(encap->from.sin_addr));
	return -1;
    }

#if 0
    printf("raw got %d bytes\n", r);
    for (i = 0; i < r; i++) {
	printf(" %02x", buf[i]);
	if ((i & 15) == 15) printf("\n");
    }
    printf("\n");
#endif

#ifdef NEED_IPID_SWAP
    p->ip_id = htons(p->ip_id);
#endif
#ifdef NEED_IPLEN_FIX
    p->ip_len = r;
#else
    p->ip_len = ntohs(p->ip_len);
#endif

    encap->buf = buf;
    encap->buflen = r;
    encap->bufpayload = (p->ip_hl << 2);
    encap->bufsize = bufsize;
    *from = encap->from;
    return r;
}

/*
 * Decapsulate from a ICMP packet
 */
int encap_icmp_recv(struct encap_method *encap,
		    unsigned char *buf, unsigned int bufsize,
		    struct sockaddr_in *from)
{
    int r;
    struct ip *p = (struct ip *)buf;
    struct icmp *ic;
    encap->fromlen = sizeof(encap->from);

    r = recvfrom(encap->fd, buf, bufsize, 0,
		    (struct sockaddr *)&encap->from, &encap->fromlen);
    if (r == -1) {
	syslog(LOG_ERR, "recvfrom: %m");
	return -1;
    }

    /*
     * Apparently, we only get packets with a correct ICMP checksum from
     * the kernel.
     */

    if ((size_t) r < (p->ip_hl << 2)+sizeof(struct icmp))
        /* Don't log (can be a reply to a ping, none of our business) */
	return -1;

    ic = (struct icmp *)(buf + (p->ip_hl << 2));
    if (ic->icmp_type != ICMP_ECHOREPLY)
        return -1;
    if (ic->icmp_code != IPPROTO_AH) /* arbitrary value */
        return -1;

    if (r < (p->ip_hl << 2)+encap->fixed_header_size) {
        syslog(LOG_ALERT, "packet too short from %s",
		inet_ntoa(encap->from.sin_addr));
	return -1;
    }

#ifdef NEED_IPID_SWAP
    p->ip_id = htons(p->ip_id);
#endif
#ifdef NEED_IPLEN_FIX
    p->ip_len = r;
#else
    p->ip_len = ntohs(p->ip_len);
#endif

    encap->buf = buf;
    encap->buflen = r;
    encap->bufpayload = (p->ip_hl << 2);
    encap->bufsize = bufsize;
    *from = encap->from;
    return r;
}

struct peer_desc *encap_udp_peer_find(struct encap_method *encap)
{
    udp_encap_header_t *eh;
    eh = (udp_encap_header_t *)encap->buf;
    return peer_find(ntohl(eh->spi), encap);
}

struct peer_desc *encap_ah_peer_find(struct encap_method *encap)
{
    ah_encap_header_t *eh;
    eh = (ah_encap_header_t *)(encap->buf + encap->bufpayload);
    return peer_find(ntohl(eh->spi), encap);
}

struct peer_desc *encap_esp_peer_find(struct encap_method *encap)
{
    esp_encap_header_t *eh;
    eh = (esp_encap_header_t *)(encap->buf + encap->bufpayload);
    return peer_find(ntohl(eh->spi), encap);
}

struct peer_desc *encap_icmp_peer_find(struct encap_method *encap)
{
    icmp_encap_header_t *eh;
    eh = (icmp_encap_header_t *)(encap->buf + encap->bufpayload);
    return peer_find(ntohl(eh->spi), encap);
}

/*
 * Decapsulate packet
 */
int encap_any_decap(struct encap_method *encap, int fd)
{
    encap->buflen
      -= encap->bufpayload+encap->fixed_header_size+encap->var_header_size;
    encap->buf
      += encap->bufpayload+encap->fixed_header_size+encap->var_header_size;
    if (encap->buflen == 0)
        return 0;
    return 1;
}

void tun_new(struct tun_method *this,
	     unsigned link_header_size, unsigned char *link_header)
{
    this->link_header_size = link_header_size;
    this->link_header = link_header;
}

/*
 * Send decapsulated packet to tunnel device
 */
int tun_send_ip(struct tun_method *this, struct encap_method *encap, int fd)
{
    int sent;

    if (this->link_header_size) {
        encap->buflen += this->link_header_size;
        encap->buf -= this->link_header_size;
        memcpy(encap->buf, this->link_header, this->link_header_size);
    }
    sent = write(fd, encap->buf, encap->buflen);
    if (sent != encap->buflen)
        syslog(LOG_ERR, "truncated in: %d -> %d\n", encap->buflen, sent);
    return 1;
}

/*
 * Initialize encap_method structures for each method
 */
int encap_udp_new(struct encap_method *encap, unsigned short port)
{
    int i;
    struct sockaddr_in source;

    encap->fd = socket(PF_INET, SOCK_DGRAM, 0);
    if (encap->fd == -1) {
	perror("socket(SOCK_DGRAM)");
	return -1;
    }

    memset(&source, 0, sizeof source);
    source.sin_addr.s_addr = INADDR_ANY;
    source.sin_port = htons(port);
    source.sin_family = AF_INET;
#ifdef HAVE_SA_LEN
    source.sin_len = sizeof(&source);
#endif

    i = bind(encap->fd, (struct sockaddr *)&source, sizeof(source));
    if (i == -1) {
	syslog(LOG_ERR, "bind: %m");
	return -1;
    }
    encap->name = "udp";
    encap->recv = encap_udp_recv;
    encap->hmac_compute = hmac_udp_compute;
    encap->peer_find = encap_udp_peer_find;
    encap->send_peer = encap_udp_send_peer;
    encap->recv_peer = encap_hmac_recv_peer;
    encap->fixed_header_size = sizeof(udp_encap_header_t);
    encap->var_header_size = 0;
    return 0;
}

int encap_ah_new(struct encap_method *encap, unsigned char proto)
{
#ifdef IP_HDRINCL
    int hincl = 1;
#endif

    encap->fd = socket(PF_INET, SOCK_RAW, proto);

    if (encap->fd == -1) {
	perror("socket(SOCK_RAW)");
	return -1;
    }
#ifdef IP_HDRINCL
    if (setsockopt(encap->fd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl))
	== -1) {
        perror("setsockopt(IP_HDRINCL)");
	close(encap->fd);
	return -1;
    }
#endif
    encap->name = "ipah";
    encap->recv = encap_rawip_recv;
    encap->hmac_compute = hmac_rawip_compute;
    encap->peer_find = encap_ah_peer_find;
    encap->send_peer = encap_ah_send_peer;
    encap->recv_peer = encap_hmac_recv_peer;
    encap->fixed_header_size = sizeof(ah_encap_header_t);
    encap->var_header_size = 0;
    return 0;
}

int encap_esp_new(struct encap_method *encap, unsigned char proto)
{
#ifdef IP_HDRINCL
    int hincl = 1;
#endif

    encap->fd = socket(PF_INET, SOCK_RAW, proto);

    if (encap->fd == -1) {
	perror("socket(SOCK_RAW)");
	return -1;
    }
#ifdef IP_HDRINCL
    if (setsockopt(encap->fd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl))
	== -1) {
        perror("setsockopt(IP_HDRINCL)");
	close(encap->fd);
	return -1;
    }
#endif
    encap->name = "ipesp";
    encap->recv = encap_rawip_recv;
    encap->hmac_compute = NULL;
    encap->peer_find = encap_esp_peer_find;
    encap->send_peer = encap_esp_send_peer;
    encap->recv_peer = encap_esp_recv_peer;
    encap->fixed_header_size = sizeof(esp_encap_header_t);
    encap->var_header_size = 0;
    return 0;
}

int encap_icmp_new(struct encap_method *encap, unsigned char proto)
{
#ifdef IP_HDRINCL
    int hincl = 1;
#endif

    encap->fd = socket(PF_INET, SOCK_RAW, proto);

    if (encap->fd == -1) {
	perror("socket(SOCK_RAW)");
	return -1;
    }
#ifdef IP_HDRINCL
    if (setsockopt(encap->fd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl))
	== -1) {
        perror("setsockopt(IP_HDRINCL)");
	close(encap->fd);
	return -1;
    }
#endif
    encap->name = "icmp";
    encap->recv = encap_icmp_recv;
    encap->hmac_compute = hmac_icmp_compute;
    encap->peer_find = encap_icmp_peer_find;
    encap->send_peer = encap_icmp_send_peer;
    encap->recv_peer = encap_hmac_recv_peer;
    encap->fixed_header_size = sizeof(icmp_encap_header_t);
    encap->var_header_size = 0;
    return 0;
}


/*
 * This is a hack to retrieve which local IP address the system would use
 * as a source when sending packets to a given destination.
 */
int find_local_addr(struct sockaddr_in *dest, struct sockaddr_in *source)
{
    int addrlen;
    struct sockaddr_in dest_socket;
    int fd;

    fd = socket(PF_INET, SOCK_DGRAM, 0);
    if (fd == -1) {
	syslog(LOG_ERR, "socket: %m");
	return -1;
    }

    memset(&dest_socket, 0, sizeof(dest_socket));

    dest_socket.sin_family = AF_INET;
#ifdef HAVE_SA_LEN
    dest_socket.sin_len = sizeof(dest_socket);
#endif
    dest_socket.sin_addr = dest->sin_addr;
    dest_socket.sin_port = htons(4444);

    if (connect(fd,
		(struct sockaddr *)&dest_socket, sizeof(dest_socket)) == -1) {
	syslog(LOG_ERR, "connect: %m");
	close(fd);
	return -1;
    }

    addrlen = sizeof(*source);

    if (getsockname(fd, (struct sockaddr *)source, &addrlen) == -1) {
	syslog(LOG_ERR, "getsockname: %m");
	close(fd);
	return -1;
    }
    close(fd);
    return 0;
}

/*
 * Retrieve and possibly update our local address to a given remote SA.
 * Return 1 if changed, 0 if not, -1 if error.
 */
int update_sa_addr(struct sa_desc *p)
{
    struct sockaddr_in new_addr;

    if (find_local_addr(&p->dest, &new_addr) == -1) {
	syslog(LOG_ALERT,
		"can't find a local address for packets to %s",
		inet_ntoa(p->dest.sin_addr));
	return -1;
    }
    if (new_addr.sin_addr.s_addr != p->source.sin_addr.s_addr) {
	char addr1[16];
	p->source.sin_addr = new_addr.sin_addr;
	strcpy(addr1, inet_ntoa(p->dest.sin_addr));
	syslog(LOG_NOTICE,
		"local address for %s is %s", addr1,
		inet_ntoa(p->source.sin_addr));
	return 1;
    }
    return 0;
}

/*
 * Find the peer record associated with a given local SPI.
 */
struct peer_desc *peer_find(unsigned long spi, struct encap_method *encap)
{
    unsigned short i;
    for (i = 0; i < peer_num; i++) {
	if (peers[i].local_sa->spi == spi
	    && peers[i].local_sa->em == encap)
	    return peers+i;
    }
    syslog(LOG_ALERT, "unknown spi %ld", spi);
    return NULL;
}

/*
 * Find the SA record for a given local SPI.
 */
struct sa_desc *find_local_sa(unsigned long spi, struct encap_method *encap)
{
    struct sa_desc *sap;
    for (sap = local_sa_list; sap; sap = sap->next)
	if (sap->spi == spi && (encap == NULL || sap->em == encap))
	    return sap;
    syslog(LOG_ALERT, "Unknown local SPI %ld", spi);
    return NULL;
}

/*
 * Find the SA record for a given remote SPI.
 */
struct sa_desc *find_remote_sa(unsigned long spi, struct encap_method *encap)
{
    struct sa_desc *sap;
    for (sap = remote_sa_list; sap; sap = sap->next)
	if (sap->spi == spi && (encap == NULL || sap->em == encap))
	    return sap;
    syslog(LOG_ALERT, "Unknown remote SPI %ld", spi);
    return NULL;
}

/*
 * Compute HMAC for an arbitrary stream of bytes
 */
int hmac_compute(EVP_MD *hm,
		 unsigned char *data, unsigned int data_size,
		 unsigned char *digest, unsigned char do_store,
		 unsigned char *secret, unsigned short secret_size)
{
    unsigned char hmac_digest[EVP_MAX_MD_SIZE];

    /* See RFC 2104 */

    HMAC (hm, secret, secret_size, data, data_size, hmac_digest, NULL);

    if (do_store) {
        memcpy(digest, hmac_digest, HMAC_SIZE);
        return 0;
    } else
        return memcmp(digest, hmac_digest, HMAC_SIZE);
}

/*
 * Compute HMAC for UDP encapsulation.
 * XXX: not really HMAC yet...
 */
int hmac_udp_compute(struct encap_method *encap,
		     unsigned char do_store,
		     struct in_addr *src_ip, EVP_MD *hm,
		     unsigned char *secret, unsigned short secret_size)
{
    EVP_MD_CTX hash_ctx;
    unsigned char hmac_digest[EVP_MAX_MD_SIZE];
    udp_encap_header_t *eh = (udp_encap_header_t *)encap->buf;

    EVP_DigestInit (&hash_ctx, hm);
    EVP_DigestUpdate(&hash_ctx, secret, secret_size);
    EVP_DigestUpdate(&hash_ctx, (unsigned char *)src_ip, sizeof(*src_ip));
    EVP_DigestUpdate(&hash_ctx, (unsigned char *)&eh->spi, sizeof(eh->spi));
    EVP_DigestUpdate(&hash_ctx,
		    (unsigned char *)&eh->seq_id, sizeof(eh->seq_id));
    EVP_DigestUpdate(&hash_ctx,
	      encap->buf+encap->fixed_header_size,
	      encap->buflen-encap->fixed_header_size);
    EVP_DigestUpdate(&hash_ctx, ":::", 3);
    EVP_DigestFinal(&hash_ctx, hmac_digest, NULL);

    if (do_store) {
        memcpy(eh->hmac, hmac_digest, HMAC_SIZE);
        return 0;
    } else
        return memcmp(eh->hmac, hmac_digest, HMAC_SIZE);
}

/*
 * Compute HMAC for ICMP encapsulation.
 * XXX: could be merged with hmac_rawip_compute
 */
int hmac_icmp_compute(struct encap_method *encap,
		      unsigned char do_store,
		      struct in_addr *src_ip, EVP_MD *hm,
		      unsigned char *secret, unsigned short secret_size)
{
    unsigned char hmac_save[HMAC_SIZE];
    icmp_encap_header_t *eh;
    struct ip *ip;
    struct icmp *icmp;
    int retval;

    unsigned char sip_ttl, sip_tos;
    unsigned short sip_sum, sip_off, sicmp_sum;

    ip = (struct ip *)encap->buf;
    icmp = (struct icmp *)(ip + 1);
    eh = (icmp_encap_header_t *)icmp;

    /* Save, then zero-out mutable fields before computing */
    sip_ttl = ip->ip_ttl;
    sip_sum = ip->ip_sum;
    sip_tos = ip->ip_tos;
    sip_off = ip->ip_off;
    sicmp_sum = icmp->icmp_cksum;
    if (!do_store)
        memcpy(hmac_save, eh->hmac, sizeof(eh->hmac));
    ip->ip_ttl = 0;
    ip->ip_sum = 0;
    ip->ip_tos = 0;
    ip->ip_off = 0;
    ip->ip_len = htons(ip->ip_len);
    memset(eh->hmac, 0, sizeof(eh->hmac));
    icmp->icmp_cksum = 0;

#if 0
    printf("computing HMAC on:\n");
    for (i = 0; i < encap->buflen; i++) {
	printf(" %02x", encap->buf[i]);
	if ((i & 15) == 15) printf("\n");
    }
    printf("\n");
#endif

    retval = hmac_compute(hm, encap->buf, encap->buflen,
			  do_store?eh->hmac:hmac_save,
			  do_store, secret, secret_size);

    /* Restore mutable fields */
    ip->ip_ttl = sip_ttl;
    ip->ip_sum = sip_sum;
    ip->ip_tos = sip_tos;
    ip->ip_off = sip_off;
    ip->ip_len = ntohs(ip->ip_len);
    icmp->icmp_cksum = sicmp_sum;

    if (!do_store)
        memcpy(eh->hmac, hmac_save, sizeof(eh->hmac));

    return retval;
}

/*
 * Compute HMAC for a IP_AH packet.
 */
int hmac_rawip_compute(struct encap_method *encap,
		       unsigned char do_store,
		       struct in_addr *src_ip, EVP_MD *hm,
		       unsigned char *secret, unsigned short secret_size)
{
    unsigned char hmac_save[HMAC_SIZE];
    ah_encap_header_t *eh;
    struct ip *ip;
    int retval;

    unsigned char sip_ttl, sip_tos;
    unsigned short sip_sum, sip_off;

    eh = (ah_encap_header_t *)(encap->buf + encap->bufpayload);
    ip = (struct ip *)encap->buf;

    /* Save, then zero-out mutable fields before computing */
    sip_ttl = ip->ip_ttl;
    sip_sum = ip->ip_sum;
    sip_tos = ip->ip_tos;
    sip_off = ip->ip_off;
    if (!do_store)
        memcpy(hmac_save, eh->hmac, sizeof(eh->hmac));
    ip->ip_ttl = 0;
    ip->ip_sum = 0;
    ip->ip_tos = 0;
    ip->ip_off = 0;
    ip->ip_len = htons(ip->ip_len);
    memset(eh->hmac, 0, sizeof(eh->hmac));

#if 0
    printf("computing HMAC on:\n");
    for (i = 0; i < encap->buflen; i++) {
	printf(" %02x", encap->buf[i]);
	if ((i & 15) == 15) printf("\n");
    }
    printf("\n");
#endif

    retval = hmac_compute(hm, encap->buf, encap->buflen,
			  do_store?eh->hmac:hmac_save,
			  do_store, secret, secret_size);

    /* Restore mutable fields */
    ip->ip_ttl = sip_ttl;
    ip->ip_sum = sip_sum;
    ip->ip_tos = sip_tos;
    ip->ip_off = sip_off;
    ip->ip_len = ntohs(ip->ip_len);

    if (!do_store)
        memcpy(eh->hmac, hmac_save, sizeof(eh->hmac));

    return retval;
}

/*
 * Encapsulate a packet in UDP and send to the peer.
 * "buf" should have exactly MAX_HEADER free bytes at its beginning
 * to account for encapsulation data (not counted in "size").
 */
void encap_udp_send_peer(struct encap_method *encap,
			 struct peer_desc *peer,
			 unsigned char *buf, unsigned int bufsize)
{
    int sent;
    udp_encap_header_t *eh;

    /* Prepend our encapsulation header */
    encap->buf = buf + MAX_HEADER - encap->fixed_header_size;
    encap->buflen = bufsize + encap->fixed_header_size;

    eh = (udp_encap_header_t *)encap->buf;
    eh->spi = ntohl(peer->remote_sa->spi);
    eh->seq_id = htonl(++peer->remote_sa->seq_id);

    hmac_udp_compute(encap, 1,
		     &peer->remote_sa->source.sin_addr, peer->remote_sa->hm,
		     peer->remote_sa->auth_secret,
		     peer->remote_sa->auth_secret_size);

    sent = sendto(encap->fd, encap->buf, encap->buflen, 0,
		  (struct sockaddr *)&peer->remote_sa->dest,
		  sizeof(peer->remote_sa->dest));
    if (sent == -1) {
	syslog(LOG_ERR, "sendto: %m");
	return;
    }
    if (sent != encap->buflen)
	syslog(LOG_ALERT, "truncated out (%d out of %d)",
		sent, encap->buflen);
}

/*
 * Encapsulate a packet in IP AH and send to the peer.
 * "buf" should have exactly MAX_HEADER free bytes at its beginning
 * to account for encapsulation data (not counted in "size").
 */
void encap_ah_send_peer(struct encap_method *encap,
			struct peer_desc *peer,
			unsigned char *buf, unsigned int bufsize)
{
    int sent;
    ah_encap_header_t *eh;
    struct ip *tip, *ip;

    buf += MAX_HEADER;

    /* Keep a pointer to the old IP header */
    tip = (struct ip *)buf;

    encap->buf = buf;
    encap->buflen = bufsize;

    /* Prepend our encapsulation header and new IP header */
    encap->buf -= encap->fixed_header_size + sizeof(struct ip);
    encap->buflen += encap->fixed_header_size + sizeof(struct ip);

    encap->bufpayload = sizeof(struct ip);

    eh = (ah_encap_header_t *)(encap->buf + encap->bufpayload);
    ip = (struct ip *)(encap->buf);

    eh->spi = htonl(peer->remote_sa->spi);
    eh->seq_id = htonl(++peer->remote_sa->seq_id);
    eh->next_header = IPPROTO_IPIP;
    eh->payload_len = 4;
    eh->reserved = 0;

    /* Fill non-mutable fields */
    ip->ip_v = IPVERSION;
    ip->ip_hl = 5;
    ip->ip_len = encap->buflen;
    ip->ip_id = htons(ip_id++);
    ip->ip_p = IPPROTO_AH;
    ip->ip_src = peer->remote_sa->source.sin_addr;
    ip->ip_dst = peer->remote_sa->dest.sin_addr;

    hmac_rawip_compute(encap, 1,
		       &peer->remote_sa->source.sin_addr,
		       peer->remote_sa->hm,
		       peer->remote_sa->auth_secret,
		       peer->remote_sa->auth_secret_size);

    /* Fill mutable fields */
    ip->ip_tos = (bufsize < sizeof(struct ip)) ? 0 : tip->ip_tos;
    ip->ip_off = 0;
    ip->ip_ttl = IPDEFTTL;
    ip->ip_sum = 0;

    sent = sendto(encap->fd, encap->buf, encap->buflen, 0,
		  (struct sockaddr *)&peer->remote_sa->dest,
		  sizeof(peer->remote_sa->dest));
    if (sent == -1) {
	syslog(LOG_ERR, "sendto: %m");
	return;
    }
    if (sent != encap->buflen)
	syslog(LOG_ALERT, "truncated out (%d out of %d)",
		sent, encap->buflen);
}

/*
 * Encapsulate a packet in IP ESP and send to the peer.
 * "buf" should have exactly MAX_HEADER free bytes at its beginning
 * to account for encapsulation data (not counted in "size").
 */
void encap_esp_send_peer(struct encap_method *encap,
			 struct peer_desc *peer,
			 unsigned char *buf, unsigned int bufsize)
{
    int sent;
    esp_encap_header_t *eh;
    struct ip *tip, *ip;
    unsigned char *iv, *cleartext;
    int i, padding;
    unsigned int cleartextlen;
    unsigned char impl_iv[EVP_MAX_IV_LENGTH];

    buf += MAX_HEADER;

    /* Keep a pointer to the old IP header */
    tip = (struct ip *)buf;

    encap->buf = buf;
    encap->buflen = bufsize;

    /*
     * Add padding as necessary
     *
     * XXX: this should be checked, RFC 2406 section 2.4 is quite
     *	    obscure on that point.
     */
    padding = (EVP_CIPHER_CTX_block_size (&peer->remote_sa->enc_key)
	       - ((encap->buflen+2) 
		  % EVP_CIPHER_CTX_block_size (&peer->remote_sa->enc_key)));
    if (padding == EVP_CIPHER_CTX_block_size (&peer->remote_sa->enc_key))
        padding = 0;

    if ((encap->buflen+2+padding) & 3)
        /*
	 * More padding needed to align padlen and next_header on
	 * the rightmost 2 bytes of a long.
	 */
        padding += 4 - ((encap->buflen+2+padding) & 3);

    for (i = 1; i <= padding; i++) {
        encap->buf[encap->buflen] = i;
	encap->buflen++;
    }

    /* Add trailing padlen and next_header */
    encap->buf[encap->buflen++] = padding;
    encap->buf[encap->buflen++] = IPPROTO_IPIP;

    cleartext = buf;
    cleartextlen = encap->buflen;

    /* Prepend our encapsulation header and new IP header */
    if (peer->remote_sa->no_iv)
	encap->var_header_size = encap->fixed_header_size;
    else
	encap->var_header_size = 
	  (encap->fixed_header_size
	   + EVP_CIPHER_CTX_iv_length (&peer->remote_sa->enc_key));

    encap->buf -= sizeof(struct ip) + encap->var_header_size;
    encap->buflen += sizeof(struct ip) + encap->var_header_size;

    encap->bufpayload = sizeof(struct ip);

    eh = (esp_encap_header_t *)(encap->buf + encap->bufpayload);
    ip = (struct ip *)(encap->buf);
    eh->spi = htonl(peer->remote_sa->spi);
    eh->seq_id = htonl(++peer->remote_sa->seq_id);

    if (peer->remote_sa->no_iv) {
	/* Implicit IV (OpenBSD compatibility mode) */
	iv = impl_iv;
	memcpy(iv, &eh->seq_id, sizeof(eh->seq_id));
	iv[4] = ~iv[0];
	iv[5] = ~iv[1];
	iv[6] = ~iv[2];
	iv[7] = ~iv[3];
    } else {
	/* Copy initialization vector in packet */
	iv = (unsigned char *)(eh + 1);
	memcpy(iv, global_iv, EVP_MAX_IV_LENGTH);
    }

    /* Fill non-mutable fields */
    ip->ip_v = IPVERSION;
    ip->ip_hl = 5;
    ip->ip_len = encap->buflen + (peer->remote_sa->hm?HMAC_SIZE:0);
    ip->ip_id = htons(ip_id++);
    ip->ip_p = IPPROTO_ESP;
    ip->ip_src = peer->remote_sa->source.sin_addr;
    ip->ip_dst = peer->remote_sa->dest.sin_addr;

    /* Fill mutable fields */
    ip->ip_tos = (bufsize < sizeof(struct ip)) ? 0 : tip->ip_tos;
    ip->ip_off = 0;
    ip->ip_ttl = IPDEFTTL;
    ip->ip_sum = 0;

#if 0
    printf("sending ESP packet (before crypt %d):\n", cleartextlen);
    for (i = 0; i < encap->buflen; i++)
        printf(" %02x", encap->buf[i]);
    printf("\n");
#endif
    
    {
      int outl;
      EVP_EncryptInit (&peer->remote_sa->enc_key, NULL, NULL, global_iv);
      EVP_EncryptUpdate (&peer->remote_sa->enc_key,
			 cleartext, &outl, cleartext, cleartextlen);
      memcpy (global_iv, peer->remote_sa->enc_key.iv, EVP_MAX_IV_LENGTH);
    }

#if 0
    printf("sending ESP packet (after crypt %d):\n", cleartextlen);
    for (i = 0; i < encap->buflen; i++)
        printf(" %02x", encap->buf[i]);
    printf("\n");
#endif

    /* Handle optional authentication field */
    if (peer->remote_sa->hm) {
#if 0
	printf("sending ESP packet (before ah %d):\n", encap->buflen);
	for (i = 0; i < encap->buflen; i++)
	    printf(" %02x", encap->buf[i]);
	printf("\n");
#endif
	hmac_compute(peer->remote_sa->hm,
		     encap->buf+encap->bufpayload,
		     encap->var_header_size+cleartextlen,
		     encap->buf+encap->bufpayload
		     +encap->var_header_size+cleartextlen,
		     1,
		     peer->remote_sa->auth_secret,
		     peer->remote_sa->auth_secret_size);
	encap->buflen += HMAC_SIZE;
#if 0
	printf("sending ESP packet (after ah %d):\n", encap->buflen);
	for (i = 0; i < encap->buflen; i++)
	    printf(" %02x", encap->buf[i]);
	printf("\n");
#endif
    }

    sent = sendto(encap->fd, encap->buf, encap->buflen, 0,
		  (struct sockaddr *)&peer->remote_sa->dest,
		  sizeof(peer->remote_sa->dest));
    if (sent == -1) {
	syslog(LOG_ERR, "sendto: %m");
	return;
    }
    if (sent != encap->buflen)
	syslog(LOG_ALERT, "truncated out (%d out of %d)",
		sent, encap->buflen);
}

/*
 * Encapsulate a packet in ICMP and send to the peer.
 * "buf" should have exactly MAX_HEADER free bytes at its beginning
 * to account for encapsulation data (not counted in "size").
 */
void encap_icmp_send_peer(struct encap_method *encap,
			  struct peer_desc *peer,
			  unsigned char *buf, unsigned int bufsize)
{
    int sent;
    icmp_encap_header_t *eh;
    struct ip *tip, *ip;

    buf += MAX_HEADER;

    /* Keep a pointer to the old IP header */
    tip = (struct ip *)buf;

    encap->buf = buf;
    encap->buflen = bufsize;

    /* Prepend our encapsulation header and new IP header */
    encap->buf -= encap->fixed_header_size + sizeof(struct ip);
    encap->buflen += encap->fixed_header_size + sizeof(struct ip);

    encap->bufpayload = sizeof(struct ip);

    eh = (icmp_encap_header_t *)(encap->buf + encap->bufpayload);
    ip = (struct ip *)(encap->buf);

    /* Fill non-mutable fields */
    ip->ip_v = IPVERSION;
    ip->ip_hl = 5;
    ip->ip_len = encap->buflen;
    ip->ip_id = htons(ip_id++);
    ip->ip_p = IPPROTO_ICMP;
    ip->ip_src = peer->remote_sa->source.sin_addr;
    ip->ip_dst = peer->remote_sa->dest.sin_addr;

    eh->icmp_type = ICMP_ECHOREPLY;
    eh->icmp_code = IPPROTO_AH; /* arbitrary */
    eh->icmp_cksum = 0;
    eh->spi = htonl(peer->remote_sa->spi);
    eh->seq_id = htonl(++peer->remote_sa->seq_id);

    hmac_icmp_compute(encap, 1,
		      &peer->remote_sa->source.sin_addr,
		      peer->remote_sa->hm,
		      peer->remote_sa->auth_secret,
		      peer->remote_sa->auth_secret_size);

    /* Fill mutable fields */
    ip->ip_tos = (bufsize < sizeof(struct ip)) ? 0 : tip->ip_tos;
    ip->ip_off = 0;
    ip->ip_ttl = IPDEFTTL;
    ip->ip_sum = 0;

    /* Fill ICMP checksum last */
    eh->icmp_cksum = in_cksum(encap->buf+sizeof(struct ip),
			      encap->buflen-sizeof(struct ip));

    sent = sendto(encap->fd, encap->buf, encap->buflen, 0,
		  (struct sockaddr *)&peer->remote_sa->dest,
		  sizeof(peer->remote_sa->dest));
    if (sent == -1) {
	syslog(LOG_ERR, "sendto: %m");
	return;
    }
    if (sent != encap->buflen)
	syslog(LOG_ALERT, "truncated out (%d out of %d)",
		sent, encap->buflen);
}

int encap_hmac_recv_peer(struct encap_method *encap,
			 struct in_addr *src_ip,
			 struct peer_desc *peer)
{
    if (encap_hmac_cmp(encap, src_ip,
		       peer->local_sa->hm,
		       peer->local_sa->auth_secret,
		       peer->local_sa->auth_secret_size) != 0) {
	syslog(LOG_ALERT, "HMAC mismatch from %s",
		inet_ntoa(*src_ip));
	return -1;
    }
    return 0;
}

int encap_esp_recv_peer(struct encap_method *encap,
			struct in_addr *src_ip,
			struct peer_desc *peer)
{
    int len, i;
    unsigned char padlen, next_header;
    unsigned char *pad;
    unsigned char impl_iv[EVP_MAX_IV_LENGTH], *iv;
    struct esp_encap_header *eh;

    eh = (struct esp_encap_header *)(encap->buf + encap->bufpayload);
    if (peer->local_sa->no_iv) {
	encap->var_header_size = 0;
	iv = impl_iv;
	memcpy(iv, &eh->seq_id, sizeof(eh->seq_id));
	iv[4] = ~iv[0];
	iv[5] = ~iv[1];
	iv[6] = ~iv[2];
	iv[7] = ~iv[3];
    } else {
	encap->var_header_size = 
	  EVP_CIPHER_CTX_iv_length (&peer->local_sa->enc_key);
	iv = encap->buf + encap->bufpayload + encap->fixed_header_size;
    }

    len = encap->buflen - encap->bufpayload
        - encap->fixed_header_size - encap->var_header_size;

    if (len < 0) {
        syslog(LOG_ALERT, "Packet too short");
        return -1;
    }

    /* Handle optional authentication field */
    if (peer->local_sa->hm) {
	len -= HMAC_SIZE;
	if (hmac_compute(peer->local_sa->hm,
			 encap->buf+encap->bufpayload,
			 encap->fixed_header_size+encap->var_header_size+len,
			 encap->buf+encap->bufpayload
			 +encap->fixed_header_size+encap->var_header_size+len,
			 0,
			 peer->local_sa->auth_secret,
			 peer->local_sa->auth_secret_size) != 0) {
	    syslog(LOG_ALERT, "HMAC mismatch in ESP mode");
	    return -1;
	}
    }

    if ((len % EVP_CIPHER_CTX_block_size (&peer->local_sa->enc_key)) != 0) {
        syslog(LOG_ALERT,
	       "payload len %d not a multiple of algorithm block size %d",
	       len, EVP_CIPHER_CTX_block_size (&peer->local_sa->enc_key));
	return -1;
    }

#if 0
    printf("receiving ESP packet (before decrypt):\n");
    for (i = 0; i < len; i++)
        printf(" %02x", encap->buf[encap->bufpayload
				  +encap->fixed_header_size
				  +encap->var_header_size+i]);
    printf("\n");
#endif
    
    {
      int outl, outl2;
      unsigned char *data;
      
      data = (encap->buf+encap->bufpayload
	      +encap->fixed_header_size
	      +encap->var_header_size);
      EVP_DecryptInit (&peer->local_sa->enc_key, NULL, NULL, iv);
      EVP_DecryptUpdate (&peer->local_sa->enc_key, data, &outl, data, len);
      EVP_EncryptUpdate (&peer->local_sa->enc_key, data + outl, &outl2, 
			 data + outl, 0);
    }

#if 0
    printf("receiving ESP packet (after decrypt %d):\n", len);
    for (i = 0; i < len; i++)
        printf(" %02x", encap->buf[encap->bufpayload
				  +encap->fixed_header_size
				  +encap->var_header_size+i]);
    printf("\n");
#endif

    padlen = encap->buf[encap->bufpayload
		       +encap->fixed_header_size
		       +encap->var_header_size+len-2];
    next_header = encap->buf[encap->bufpayload
			    +encap->fixed_header_size
			    +encap->var_header_size+len-1];

    if (padlen+2 > len) {
        syslog(LOG_ALERT, "Inconsistent padlen");
	return -1;
    }
    if (next_header != IPPROTO_IPIP) {
        syslog(LOG_ALERT, "Inconsistent next_header %d", next_header);
	return -1;
    }

#if 0
    printf("pad len: %d, next_header: %d\n", padlen, next_header);
#endif
    len -= padlen + 2;

    /* Check padding */
    pad = encap->buf + encap->bufpayload
        + encap->fixed_header_size
        + encap->var_header_size + len;
    for (i = 1; i <= padlen; i++) {
        if (*pad != i) {
	    syslog(LOG_ALERT, "Bad padding");
	    return -1;
        }
	pad++;
    }

    return 0;
}

static void
vpnc_main_loop (struct peer_desc *peer,
		struct encap_method *meth,
		int tun_fd)
{
  struct pollfd pollfds[2];
  
  pollfds[0].fd = tun_fd;
  pollfds[0].events = POLLIN;
  pollfds[1].fd = encap_get_fd (meth);
  pollfds[1].events = POLLIN;

  for (;;)
    {
      int presult;
      
      do {
	presult = poll (pollfds, sizeof (pollfds) / sizeof (pollfds[0]), -1);
      } while (presult == -1 && errno == EINTR);
      if (presult == -1)
	{
	  syslog(LOG_ERR, "poll: %m");
	  continue;
	}

      if (pollfds[0].revents & POLLIN)
	{
	  int pack;
	  
	  /* Receive a packet from the tunnel interface */
	  pack = read(peer->tun_fd,
		      buf+MAX_HEADER-peer->tm->link_header_size,
		      MAX_PACKET+peer->tm->link_header_size);
	  if (pack == -1) {
	    syslog(LOG_ERR, "read: %m");
	    continue;
	  }
	  
	  if (peer->remote_sa->use_dest == 0) {
	    syslog(LOG_NOTICE,
		   "peer hasn't a known address yet");
	    continue;
	  }
	  
	  if (((struct ip *)(buf+MAX_HEADER))->ip_dst.s_addr
	      == peer->remote_sa->dest.sin_addr.s_addr) {
	    syslog(LOG_ALERT, "routing loop to %s",
		   inet_ntoa(peer->remote_sa->dest.sin_addr));
	    continue;
	  }
	  
	  /* Encapsulate and send to the other end of the tunnel */
	  encap_send_peer(peer->remote_sa->em, peer, buf, pack);
	  
	  /* Update sent packet timeout */
	  peer->remote_sa->last_packet_sent = time (NULL);
	}
      if (pollfds[1].revents & POLLIN)
	{
	  /* Receive a packet from a socket */
	  struct peer_desc *peer;
	  int pack;
	  struct sockaddr_in from;
	  
	  pack = encap_recv(meth, buf, MAX_HEADER+MAX_PACKET, &from);
	  if (pack == -1)
	    continue;
	  
	  peer = encap_peer_find(meth);
	  if (peer == NULL) {
	    syslog(LOG_NOTICE, "unknown spi from %s",
		   inet_ntoa(from.sin_addr));
	    continue;
	  }
	  
	  /* Check auth digest and/or decrypt */
	  if (encap_recv_peer(meth, &from.sin_addr, peer) != 0)
	    continue;
	  
	  /* Check origin IP; update our copy if need be */
	  if (peer->remote_sa->use_dest == 0
	      || from.sin_addr.s_addr
	      != peer->remote_sa->dest.sin_addr.s_addr) {
	    /* remote end changed address */
	    char addr1[16];
	    strcpy(addr1, inet_ntoa(peer->remote_sa->dest.sin_addr));
	    syslog(LOG_NOTICE,
		   "spi %ld: remote address changed from %s to %s",
		   peer->remote_sa->spi,
		   addr1,
		   inet_ntoa(from.sin_addr));
	    peer->remote_sa->dest.sin_addr.s_addr
	      = from.sin_addr.s_addr;
	    peer->remote_sa->use_dest = 1;
	    update_sa_addr(peer->remote_sa);
	  }
	  /* Update received packet timeout */
	  peer->remote_sa->last_packet_recv = time (NULL);
	  
	  if (encap_any_decap(meth, peer->tun_fd) == 0)
	    syslog(LOG_DEBUG, "received update probe from peer");
	  else
	    /* Send the decapsulated packet to the tunnel interface */
	    tun_send_ip(peer->tm, meth, peer->tun_fd);
	}
    }
}

static uint8_t *volatile kill_packet;
static size_t volatile kill_packet_size;
static struct sockaddr * volatile kill_dest;

void
killit(int signum)
{
  int sock;
  sock = socket (AF_INET, SOCK_DGRAM, 0);
  if (sock >= 0)
    {
      sendto (sock, kill_packet, kill_packet_size, 0,
	      kill_dest, sizeof (struct sockaddr_in));
      close (sock);
    }
  _exit (0);
}

void
vpnc_doit(unsigned long tous_spi,
	  const unsigned char *tous_keys, 
	  struct sockaddr_in *tous_dest,
	  unsigned long tothem_spi,
	  const unsigned char *tothem_keys,
	  struct sockaddr_in *tothem_dest,
	  int tun_fd, int random_fd, int sockfd,
	  uint8_t *kill_packet_p, size_t kill_packet_size_p,
	  struct sockaddr *kill_dest_p)
{
  struct encap_method meth;

  static struct sa_desc tous_sa, tothem_sa;
  time_t t = time(NULL);

  read (random_fd, global_iv, sizeof (global_iv));
  
  tun_new(&tm_tun, 0, NULL);
  
  if (encap_esp_new(&meth, IPPROTO_ESP) == -1)
    exit(1);

  tous_sa.next = remote_sa_list;
  remote_sa_list = &tous_sa;
  tous_sa.em = &meth;
  tous_sa.last_packet_recv = t;
  tous_sa.last_packet_sent = t;
  tous_sa.last_checkifaddr = t;
  tous_sa.hm = EVP_md5();
  tous_sa.spi = htonl (tous_spi);
  memcpy (tous_sa.auth_secret, tous_keys + 24, 16);
  tous_sa.auth_secret_size = 16;
  memcpy (tous_sa.enc_secret, tous_keys, 24);
  tous_sa.enc_secret_size = 24;
  des_set_odd_parity((des_cblock *)tous_sa.enc_secret);
  des_set_odd_parity((des_cblock *)(tous_sa.enc_secret+8));
  des_set_odd_parity((des_cblock *)(tous_sa.enc_secret+16));
  memcpy (&tous_sa.init, tous_dest, sizeof (struct sockaddr_in));
  memcpy (&tous_sa.dest, tous_dest, sizeof (struct sockaddr_in));
  if (update_sa_addr(&tous_sa) != -1)
    {
      tous_sa.use_fallback = 1;
      tous_sa.use_dest = 1;
    }
  EVP_DecryptInit (&tous_sa.enc_key, EVP_des_ede3_cbc(),
		   tous_sa.enc_secret, NULL);
  
  tothem_sa.next = local_sa_list;
  local_sa_list = &tothem_sa;
  tothem_sa.em = &meth;
  tothem_sa.last_packet_recv = t;
  tothem_sa.last_packet_sent = t;
  tothem_sa.last_checkifaddr = t;
  tothem_sa.hm = EVP_md5();
  tothem_sa.spi = htonl (tothem_spi);
  memcpy (tothem_sa.auth_secret, tothem_keys + 24, 16);
  tothem_sa.auth_secret_size = 16;
  memcpy (tothem_sa.enc_secret, tothem_keys, 24);
  tothem_sa.enc_secret_size = 24;
  des_set_odd_parity((des_cblock *)tothem_sa.enc_secret);
  des_set_odd_parity((des_cblock *)(tothem_sa.enc_secret+8));
  des_set_odd_parity((des_cblock *)(tothem_sa.enc_secret+16));
  memcpy (&tothem_sa.init, tothem_dest, sizeof (struct sockaddr_in));
  memcpy (&tothem_sa.dest, tothem_dest, sizeof (struct sockaddr_in));
  if (update_sa_addr(&tothem_sa) != -1)
    {
      tothem_sa.use_fallback = 1;
      tothem_sa.use_dest = 1;
    }
  EVP_EncryptInit (&tothem_sa.enc_key, EVP_des_ede3_cbc(),
		   tothem_sa.enc_secret, NULL);
  
  peers[peer_num].tun_fd = tun_fd;
  peers[peer_num].local_sa = &tous_sa;
  peers[peer_num].remote_sa = &tothem_sa;
  peers[peer_num].tm = &tm_tun;
  peer_num++;

  openlog ("pipsecd", LOG_PID, LOG_DAEMON);
  if (fork() == 0)
    {
      kill_packet = kill_packet_p;
      kill_packet_size = kill_packet_size_p;
      kill_dest = kill_dest_p;
      
      signal (SIGHUP, killit);
      signal (SIGINT, killit);
      signal (SIGTERM, killit);
      signal (SIGKILL, killit);
      signal (SIGXCPU, killit);
      signal (SIGPWR, killit);

      chdir ("/");
      close (0); close (1); close (2); close (random_fd);
      setsid ();

      vpnc_main_loop (&peers[peer_num - 1], &meth, tun_fd);
    }
}
