/* IPSec VPN client compatible with Cisco equipment.
   Copyright (C) 2002  Geoffrey Keating

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <sys/fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <error.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <poll.h>
#include <net/if.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>

#include <openssl/bn.h>
#include <openssl/dh.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>

#include "isakmp-pkt.h"

extern void vpnc_doit(unsigned long tous_spi,
		      const unsigned char *tous_keys, 
		      struct sockaddr_in *tous_dest,
		      unsigned long tothem_spi,
		      const unsigned char *tothem_keys,
		      struct sockaddr_in *tothem_dest,
		      int tun_fd, int random_fd, int sockfd,
		      uint8_t *kill_packet_p, size_t kill_packet_size_p,
		      struct sockaddr *kill_dest_p);

#undef TEST_VIS

static int random_fd = -1;

static void
get_random (void *buf, ssize_t cnt)
{
  if (random_fd == -1)
    {
      random_fd = open ("/dev/urandom", O_RDONLY);
      if (random_fd == -1)
	error (1, errno, "opening /dev/urandom");
    }
  if (read (random_fd, buf, cnt) != cnt)
    error (1, errno, "fetching %d random bytes", cnt);

#ifdef TEST_VIS
  {
    ssize_t i;
    printf ("random: ");
    for (i = 0; i < cnt; i++)
      printf ("0x%02x,%c", ((uint8_t *)buf)[i], i % 12 == 11 ? '\n' : ' ');
    printf ("\n");
  }
#endif
}

static int
make_socket (uint16_t port)
{
  int sock;
  struct sockaddr_in name;

  /* Create the socket. */
  sock = socket (PF_INET, SOCK_DGRAM, 0);
  if (sock < 0)
    error (1, errno, "making socket");

  /* Give the socket a name. */
  name.sin_family = AF_INET;
  name.sin_port = htons (port);
  name.sin_addr.s_addr = htonl (INADDR_ANY);
  if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
    error (1, errno, "binding to port %d", port);

  return sock;
}

static struct sockaddr *
init_sockaddr (const char *hostname,
	       uint16_t port)
{
  struct hostent *hostinfo;
  struct sockaddr_in *result;
  
  result = malloc (sizeof (struct sockaddr_in));
  if (result == NULL)
    error (1, errno, "out of memory");

  result->sin_family = AF_INET;
  result->sin_port = htons (port);
  if (inet_aton (hostname, &result->sin_addr) == 0)
    {
      hostinfo = gethostbyname (hostname);
      if (hostinfo == NULL)
	error (1, 0, "unknown host `%s'\n", hostname);
      result->sin_addr = *(struct in_addr *) hostinfo->h_addr;
    }
  return (struct sockaddr *)result;
}

static int tun_fd = -1;
static char tun_name[IFNAMSIZ];

static void
setup_tunnel(void)
{
  struct ifreq req;

  tun_fd = open ("/dev/tun", O_RDWR);
  if (tun_fd == -1)
    error (1, errno, "can't open /dev/tun");

  memset (&req, 0, sizeof (req));
  req.ifr_flags = IFF_TUN | IFF_NO_PI;

  if (ioctl (tun_fd, TUNSETIFF, &req) != 0)
    error (1, errno, "can't initialise interface");

  memcpy (tun_name, req.ifr_name, IFNAMSIZ);
}

static int sockfd = -1;
static struct sockaddr *dest_addr;
static int timeout = 5000;  /* 5 seconds */
static uint8_t resend_hash[SHA_DIGEST_LENGTH];

static int
recv_ignore_dup (void *recvbuf, size_t recvbufsize)
{
  uint8_t resend_check_hash[SHA_DIGEST_LENGTH];
  int recvsize;
  struct sockaddr_in recvaddr;
  socklen_t recvaddr_size = sizeof (recvaddr);
  char ntop_buf[32];

  recvsize = recvfrom (sockfd, recvbuf, recvbufsize, 0, &recvaddr,
		       &recvaddr_size);
  if (recvsize == -1)
    error (1, errno, "recieving packet");
  if (recvsize > 0)
    {
      if (recvaddr_size != sizeof (recvaddr)
	  || recvaddr.sin_family != dest_addr->sa_family
	  || recvaddr.sin_port != ((struct sockaddr_in *)dest_addr)->sin_port
	  || memcmp (&recvaddr.sin_addr, 
		     &((struct sockaddr_in *)dest_addr)->sin_addr, 
		     sizeof (struct in_addr)) != 0)
	{
	  error (0, 0, "got response from unknown host %s:%d",
		 inet_ntop (recvaddr.sin_family, &recvaddr.sin_addr,
			    ntop_buf, sizeof (ntop_buf)),
		 ntohs (recvaddr.sin_port));
	  return -1;
	}

      SHA1 (recvbuf, recvsize, resend_check_hash);
      if (memcmp (resend_hash, resend_check_hash, SHA_DIGEST_LENGTH)
	  == 0)
	return -1;
      memcpy (resend_hash, resend_check_hash, SHA_DIGEST_LENGTH);
    }
  return recvsize;
}

/* Send TOSEND of size SENDSIZE to the socket.  Then wait for a new packet,
   resending TOSEND on timeout, and ignoring duplicate packets; the
   new packet is put in RECVBUF of size RECVBUFSIZE and the actual size
   of the new packet is returned.  */

static ssize_t
sendrecv (void *recvbuf, size_t recvbufsize, void *tosend, size_t sendsize)
{
  struct pollfd pfd;
  int tries = 0;
  int recvsize;
  time_t start = time (NULL);
  time_t end;
  
  pfd.fd = sockfd;
  pfd.events = POLLIN;
  tries = 0;
  
  for (;;)
    {
      int pollresult;
      
      if (sendto (sockfd, tosend, sendsize, 0,
		  dest_addr, sizeof (struct sockaddr_in)) != (int) sendsize)
	error (1, errno, "can't send packet");
      do {
	pollresult = poll (&pfd, 1, timeout << tries);
      } while (pollresult == -1 && errno == EINTR);
      if (pollresult == -1)
	error (1, errno, "can't poll socket");
      if (pollresult != 0)
	{
	  recvsize = recv_ignore_dup (recvbuf, recvbufsize);
	  end = time (NULL);
	  if (recvsize != -1)
	    break;
	  continue;
	}
      if (tries > 5)
	error (1, 0, "no response from target");
      tries++;
    }

  /* Wait at least 2s for a response or 4 times the time it took
     last time.  */
  if (start == end)
    timeout = 2000;
  else
    timeout = 4000 * (end - start);

#ifdef TEST_VIS
  {
    int i;
    printf ("recv: ");
    for (i = 0; i < recvsize; i++)
      printf ("0x%02x,%c", ((uint8_t *)recvbuf)[i],
	      i % 12 == 11 ? '\n' : ' ');
    printf ("\n");
  }
#endif

  return recvsize;
}

enum config_enum {
  CONFIG_IPSEC_GATEWAY,
  CONFIG_IPSEC_ID,
  CONFIG_IPSEC_SECRET,
  CONFIG_XAUTH_USERNAME,
  CONFIG_XAUTH_PASSWORD,
  LAST_CONFIG
};

static const char *config[LAST_CONFIG];

struct isakmp_payload *
make_our_sa (void)
{
  struct isakmp_payload *r = new_isakmp_payload (ISAKMP_PAYLOAD_SA);
  struct isakmp_attribute *a;
  
  r->u.sa.doi = ISAKMP_DOI_IPSEC;
  r->u.sa.situation = ISAKMP_IPSEC_SIT_IDENTITY_ONLY;
  r->u.sa.proposals = new_isakmp_payload (ISAKMP_PAYLOAD_P);
  r->u.sa.proposals->u.p.prot_id = ISAKMP_IPSEC_PROTO_ISAKMP;
  r->u.sa.proposals->u.p.transforms = new_isakmp_payload (ISAKMP_PAYLOAD_T);
  r->u.sa.proposals->u.p.transforms->u.t.id = ISAKMP_IPSEC_KEY_IKE;
  a = new_isakmp_attribute_16 (IKE_ATTRIB_GROUP_DESC, IKE_GROUP_MODP_1024,
			       NULL);
  a = new_isakmp_attribute_16 (IKE_ATTRIB_AUTH_METHOD, 
			       XAUTH_AUTH_XAUTHInitPreShared, a);
  a = new_isakmp_attribute_16 (IKE_ATTRIB_HASH, IKE_HASH_MD5, a);
  a = new_isakmp_attribute_16 (IKE_ATTRIB_ENC, IKE_ENC_3DES_CBC, a);
  r->u.sa.proposals->u.p.transforms->u.t.attributes = a;
  return r;
}

struct sa_block 
{
  uint8_t i_cookie[ISAKMP_COOKIE_LENGTH];
  uint8_t r_cookie[ISAKMP_COOKIE_LENGTH];
  unsigned char des_key[32];
  unsigned char initial_iv[EVP_MAX_IV_LENGTH];
  unsigned char skeyid_a[EVP_MAX_MD_SIZE];
  unsigned char skeyid_d[EVP_MAX_MD_SIZE];
  EVP_CIPHER *cipher;
  EVP_MD *hash;
  uint8_t current_iv_msgid[4];
  unsigned char current_iv[EVP_MAX_IV_LENGTH];
  uint8_t our_address[4], our_netmask[4];
  uint32_t tous_esp_spi, tothem_esp_spi;
  uint8_t *kill_packet;
  size_t kill_packet_size;
};

void
isakmp_crypt (struct sa_block *s, uint8_t *block, size_t blocklen, int enc)
{
  unsigned char new_iv[EVP_MAX_IV_LENGTH];
  EVP_CIPHER_CTX c;
  int outl, outl2 = 0;
  const int ivlen = EVP_CIPHER_iv_length (s->cipher);

  if (blocklen < ISAKMP_PAYLOAD_O 
      || ((blocklen - ISAKMP_PAYLOAD_O) % EVP_CIPHER_block_size (s->cipher) 
	  != 0))
    abort ();

  if (memcmp (block + ISAKMP_MESSAGE_ID_O, s->current_iv_msgid, 4) != 0)
    {
      unsigned char iv[EVP_MAX_MD_SIZE];
      EVP_MD_CTX ctx;
      EVP_DigestInit (&ctx, s->hash);
      EVP_DigestUpdate (&ctx, s->initial_iv, ivlen);
      EVP_DigestUpdate (&ctx, block + ISAKMP_MESSAGE_ID_O, 4);
      EVP_DigestFinal (&ctx, iv, NULL);
      if (enc >= 0)
	{
	  memcpy (s->current_iv, iv, ivlen);
	  memcpy (s->current_iv_msgid, block + ISAKMP_MESSAGE_ID_O, 4);
	}
    }

  if (!enc)
    memcpy (new_iv, block + blocklen - ivlen, ivlen);
  EVP_CipherInit (&c, s->cipher, s->des_key, s->current_iv, enc != 0);
  EVP_CipherUpdate (&c, block + ISAKMP_PAYLOAD_O, &outl,
		    block + ISAKMP_PAYLOAD_O, blocklen - ISAKMP_PAYLOAD_O);
  /* There may be a cached block.  */
  if (! enc)
    EVP_EncryptUpdate (&c, block + ISAKMP_PAYLOAD_O + outl, &outl2,
		       block + ISAKMP_PAYLOAD_O + outl, 0);
  if (outl + outl2 + ISAKMP_PAYLOAD_O != (int)blocklen)
    error (3, 0, "error: %d + %d + %d != %d", outl, outl2, ISAKMP_PAYLOAD_O,
	   blocklen);
  EVP_CIPHER_CTX_cleanup (&c);
  if (enc > 0)
    memcpy (s->current_iv, block + blocklen - ivlen, ivlen);
  else if (enc == 0)
    memcpy (s->current_iv, new_iv, ivlen);

#ifdef TEST_VIS
  if (!enc) 
    {
      size_t i;
      printf ("recvD: ");
      for (i = 0; i < blocklen; i++)
	printf ("0x%02x,%c", block[i],
		i % 12 == 11 ? '\n' : ' ');
      printf ("\n");
    }
#endif
}

static uint8_t r_packet[2048];
static ssize_t r_length;

void
do_phase_1 (const char *key_id, const char *shared_key,
	    struct sa_block *d)
{
  unsigned char i_nonce[20];
  DH *dh;
  unsigned char dh_public[1024/8];
  unsigned char returned_hash[EVP_MAX_MD_SIZE];
  static const uint8_t xauth_vid[] = XAUTH_VENDOR_ID;
  static const uint8_t my_vid[] = { 
    0x35, 0x53, 0x07, 0x6c, 0x4f, 0x65, 0x12, 0x68, 0x02, 0x82, 0xf2, 0x15,
    0x8a, 0xa8, 0xa0, 0x9e };
  
  struct isakmp_packet *p1;
  
  get_random (d->i_cookie, ISAKMP_COOKIE_LENGTH);
  if (d->i_cookie[0] == 0)
    d->i_cookie[0] = 1;
  get_random (i_nonce, sizeof (i_nonce));
  /* Set up the Diffie-Hellman stuff.  */
  {
    unsigned char dh_key[1024/8];
    static const char p[] = 
      "FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD1"
      "29024E088A67CC74020BBEA63B139B22514A08798E3404DD"
      "EF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245"
      "E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7ED"
      "EE386BFB5A899FA5AE9F24117C4B1FE649286651ECE65381"
      "FFFFFFFFFFFFFFFF";

    dh = DH_new();
    BN_hex2bn (&dh->p, p);
    BN_hex2bn (&dh->g, "2");
    dh_key[0] = 0;
    get_random (dh_key + 1, sizeof (dh_key) - 1);
    dh->priv_key = BN_bin2bn (dh_key, sizeof (dh_key), NULL);
    DH_generate_key (dh);
    memset (dh_public, 0, sizeof (dh_public));
    BN_bn2bin (dh->pub_key, (dh_public + sizeof (dh_public) 
			     - BN_num_bytes (dh->pub_key)));
  }
  
  /* Create the first packet.  */
  {
    struct isakmp_payload *l;
    uint8_t *pkt;
    size_t pkt_len;

    p1 = new_isakmp_packet();
    memcpy (p1->i_cookie, d->i_cookie, ISAKMP_COOKIE_LENGTH);
    p1->isakmp_version = ISAKMP_VERSION;
    p1->exchange_type = ISAKMP_EXCHANGE_AGGRESSIVE;
    p1->payload = l = make_our_sa();
    l->next = new_isakmp_data_payload (ISAKMP_PAYLOAD_KE,
				       dh_public, sizeof (dh_public));
    l->next->next = new_isakmp_data_payload (ISAKMP_PAYLOAD_NONCE,
					     i_nonce, sizeof (i_nonce));
    l = l->next->next;
    l->next = new_isakmp_payload (ISAKMP_PAYLOAD_ID);
    l = l->next;
    l->u.id.type = ISAKMP_IPSEC_ID_KEY_ID;
    l->u.id.length = (strlen (key_id) + 1 + 3) & ~3;
    l->u.id.data = xallocc (l->u.id.length);
    memcpy (l->u.id.data, key_id, strlen (key_id));
    l->next = new_isakmp_data_payload (ISAKMP_PAYLOAD_VID,
				       xauth_vid, sizeof (xauth_vid));
    l->next->next = new_isakmp_data_payload (ISAKMP_PAYLOAD_VID,
					     my_vid, sizeof (my_vid));
    l->next->next->next = new_isakmp_data_payload (ISAKMP_PAYLOAD_VID,
						   my_vid, sizeof (my_vid));
    flatten_isakmp_packet (p1, &pkt, &pkt_len);

    /* Now, send that packet and receive a new one.  */
    r_length = sendrecv (r_packet, sizeof (r_packet), 
			 pkt, pkt_len);
    free (pkt);
  }
  /* Decode the recieved packet.  */
  {
     struct isakmp_packet *r;
     uint16_t reject;
     struct isakmp_payload *rp;
     struct isakmp_payload *nonce = NULL;
     struct isakmp_payload *ke = NULL;
     struct isakmp_payload *hash = NULL;
     struct isakmp_payload *idp = NULL;
     int seen_xauth_vid = 0;
     unsigned char skeyid[EVP_MAX_MD_SIZE];
     
     reject = 0;
     r = parse_isakmp_packet (r_packet, r_length, &reject);

     /* Verify the correctness of the recieved packet.  */
     if (reject == 0 && 
	 memcmp (r->i_cookie, d->i_cookie, ISAKMP_COOKIE_LENGTH) != 0)
       reject = ISAKMP_N_INVALID_COOKIE;
     if (reject == 0)
       memcpy (d->r_cookie, r->r_cookie, ISAKMP_COOKIE_LENGTH);
     if (reject == 0 && r->exchange_type != ISAKMP_EXCHANGE_AGGRESSIVE)
       reject = ISAKMP_N_INVALID_EXCHANGE_TYPE;
     if (reject == 0 && r->flags != 0)
       reject = ISAKMP_N_INVALID_FLAGS;
     if (reject == 0 && r->message_id != 0)
       reject = ISAKMP_N_INVALID_MESSAGE_ID;
     if (reject != 0)
       error (1, 0, "response was invalid [1]: %s", 
	      isakmp_notify_to_error (reject));
     for (rp = r->payload; rp && reject == 0; rp = rp->next)
       switch (rp->type)
	 {
	 case ISAKMP_PAYLOAD_SA:
	   if (reject == 0 && rp->u.sa.doi != ISAKMP_DOI_IPSEC)
	     reject = ISAKMP_N_DOI_NOT_SUPPORTED;
	   if (reject == 0 && 
	       rp->u.sa.situation != ISAKMP_IPSEC_SIT_IDENTITY_ONLY)
	     reject = ISAKMP_N_SITUATION_NOT_SUPPORTED;
	   if (reject == 0 && 
	       (rp->u.sa.proposals == NULL
		|| rp->u.sa.proposals->next != NULL))
	     reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
	   if (reject == 0 &&
	       rp->u.sa.proposals->u.p.prot_id != ISAKMP_IPSEC_PROTO_ISAKMP)
	     reject = ISAKMP_N_INVALID_PROTOCOL_ID;
	   if (reject == 0 &&
	       rp->u.sa.proposals->u.p.spi_size != 0)
	     reject = ISAKMP_N_INVALID_SPI;
	   if (reject == 0 &&
	       (rp->u.sa.proposals->u.p.transforms == NULL
		|| rp->u.sa.proposals->u.p.transforms->next != NULL))
	     reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
	   if (reject == 0 &&
	       (rp->u.sa.proposals->u.p.transforms->u.t.id 
		!= ISAKMP_IPSEC_KEY_IKE))
	     reject = ISAKMP_N_INVALID_TRANSFORM_ID;
	   if (reject == 0) {
	     struct isakmp_attribute *a 
	       = rp->u.sa.proposals->u.p.transforms->u.t.attributes;
	     int seen_enc = 0, seen_hash = 0, seen_auth = 0, seen_group = 0;
	     for (; a && reject == 0; a = a->next)
	       switch (a->type)
		 {
		   case IKE_ATTRIB_GROUP_DESC: 
		     if (a->af == isakmp_attr_16 &&
			 a->u.attr_16 == IKE_GROUP_MODP_1024)
		       seen_group = 1;
		     else
		       reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
		     break;
		   case IKE_ATTRIB_AUTH_METHOD: 
		     if (a->af == isakmp_attr_16 &&
			 a->u.attr_16 == XAUTH_AUTH_XAUTHInitPreShared)
		       seen_auth = 1;
		     else
		       reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
		     break;
		   case IKE_ATTRIB_HASH: 
		     if (a->af == isakmp_attr_16 &&
			 a->u.attr_16 == IKE_HASH_MD5)
		       seen_hash = 1;
		     else
		       reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
		     break;
		   case IKE_ATTRIB_ENC: 
		     if (a->af == isakmp_attr_16 &&
			 a->u.attr_16 == IKE_ENC_3DES_CBC)
		       seen_enc = 1;
		     else
		       reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
		     break;
		 default:
		   reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
		   break;
		 }
	     if (! seen_group || ! seen_auth || ! seen_hash || ! seen_enc)
	       reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
	   }
	   break;

	 case ISAKMP_PAYLOAD_ID:	idp   = rp; break;
	 case ISAKMP_PAYLOAD_KE:	ke    = rp; break;
	 case ISAKMP_PAYLOAD_NONCE:	nonce = rp; break;
	 case ISAKMP_PAYLOAD_HASH:	hash  = rp; break;
	 case ISAKMP_PAYLOAD_VID:
	   if (rp->u.vid.length == sizeof (xauth_vid)
	       && memcmp (rp->u.vid.data, xauth_vid, sizeof (xauth_vid)) == 0)
	     seen_xauth_vid = 1;
	   break;
	 default:
	   reject = ISAKMP_N_INVALID_PAYLOAD_TYPE;
	   break;
	 }

     d->cipher = EVP_des_ede3_cbc();
     d->hash = EVP_md5();

     if (reject == 0
	 && (ke == NULL || ke->u.ke.length != sizeof (dh_public)))
       reject = ISAKMP_N_INVALID_KEY_INFORMATION;
     if (reject == 0 && nonce == NULL)
       reject = ISAKMP_N_INVALID_HASH_INFORMATION;
     if (reject != 0)
       error (1, 0, "response was invalid [2]: %s", 
	      isakmp_notify_to_error (reject));
     if (reject == 0 && idp == NULL)
       reject = ISAKMP_N_INVALID_ID_INFORMATION;
     if (reject == 0 && (hash == NULL 
			 || hash->u.hash.length != EVP_MD_size(d->hash)))
       reject = ISAKMP_N_INVALID_HASH_INFORMATION;
     if (reject != 0)
       error (1, 0, "response was invalid [3]: %s", 
	      isakmp_notify_to_error (reject));

     /* Generate SKEYID.  */
     {
       HMAC_CTX hm;
       unsigned int hlen;

       HMAC_Init (&hm, shared_key, strlen (shared_key), d->hash);
       HMAC_Update (&hm, i_nonce, sizeof (i_nonce));
       HMAC_Update (&hm, nonce->u.nonce.data, nonce->u.nonce.length);
       HMAC_Final (&hm, skeyid, &hlen);
       HMAC_cleanup (&hm);
     }

     /* Verify the hash.  */
     {
       HMAC_CTX hm;
       unsigned char expected_hash[EVP_MAX_MD_SIZE];
       uint8_t *sa_f, *idi_f, *idp_f;
       size_t sa_size, idi_size, idp_size;
       struct isakmp_payload *sa, *idi;

       sa = p1->payload;
       for (idi = sa; idi->type != ISAKMP_PAYLOAD_ID; idi = idi->next)
	 ;
       sa->next = NULL;
       idi->next = NULL;
       idp->next = NULL;
       flatten_isakmp_payload (sa, &sa_f, &sa_size);
       flatten_isakmp_payload (idi, &idi_f, &idi_size);
       flatten_isakmp_payload (idp, &idp_f, &idp_size);

       HMAC_Init (&hm, skeyid, EVP_MD_size (d->hash), d->hash);
       HMAC_Update (&hm, ke->u.ke.data, ke->u.ke.length);
       HMAC_Update (&hm, dh_public, sizeof (dh_public));
       HMAC_Update (&hm, d->r_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, d->i_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, sa_f + 4, sa_size - 4);
       HMAC_Update (&hm, idp_f + 4, idp_size - 4);
       HMAC_Final (&hm, expected_hash, NULL);
       HMAC_cleanup (&hm);

       if (memcmp (expected_hash, hash->u.hash.data, 
		   EVP_MD_size (d->hash)) != 0)
	 {
	   error (1, 0, "hash comparison failed: %s", 
		  isakmp_notify_to_error (ISAKMP_N_AUTHENTICATION_FAILED));
	 }

       HMAC_Init (&hm, skeyid, EVP_MD_size (d->hash), d->hash);
       HMAC_Update (&hm, dh_public, sizeof (dh_public));
       HMAC_Update (&hm, ke->u.ke.data, ke->u.ke.length);
       HMAC_Update (&hm, d->i_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, d->r_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, sa_f + 4, sa_size - 4);
       HMAC_Update (&hm, idi_f + 4, idi_size - 4);
       HMAC_Final (&hm, returned_hash, NULL);
       HMAC_cleanup (&hm);

       free (sa_f);
       free (idi);
       free (idp);
     }

     /* Determine all the SKEYID_x keys.  */
     {
       HMAC_CTX hm;
       static const unsigned char c012[3] = { 0, 1, 2 };
       unsigned char skeyid_e[EVP_MAX_MD_SIZE];
       unsigned char dh_shared_secret[1024/8];

       /* Determine the shared secret.  */
       {
	 BIGNUM *other_public;
	 
	 other_public = BN_bin2bn (ke->u.ke.data, ke->u.ke.length, NULL);
	 DH_compute_key (dh_shared_secret, other_public, dh);
	 BN_free (other_public);
       }

       HMAC_Init (&hm, skeyid, EVP_MD_size (d->hash), d->hash);
       HMAC_Update (&hm, dh_shared_secret, sizeof (dh_shared_secret));
       HMAC_Update (&hm, d->i_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, d->r_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, c012+0, 1);
       HMAC_Final (&hm, d->skeyid_d, NULL);
       HMAC_Init (&hm, skeyid, EVP_MD_size (d->hash), d->hash);
       HMAC_Update (&hm, d->skeyid_d, EVP_MD_size (d->hash));
       HMAC_Update (&hm, dh_shared_secret, sizeof (dh_shared_secret));
       HMAC_Update (&hm, d->i_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, d->r_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, c012+1, 1);
       HMAC_Final (&hm, d->skeyid_a, NULL);
       HMAC_Init (&hm, skeyid, EVP_MD_size (d->hash), d->hash);
       HMAC_Update (&hm, d->skeyid_a, EVP_MD_size (d->hash));
       HMAC_Update (&hm, dh_shared_secret, sizeof (dh_shared_secret));
       HMAC_Update (&hm, d->i_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, d->r_cookie, ISAKMP_COOKIE_LENGTH);
       HMAC_Update (&hm, c012+2, 1);
       HMAC_Final (&hm, skeyid_e, NULL);
       HMAC_cleanup (&hm);
       memset (dh_shared_secret, 0, sizeof (dh_shared_secret));

       /* Determine the 3DES CBC keys.  */
       HMAC (d->hash, skeyid_e, EVP_MD_size (d->hash), c012,
	     1, d->des_key, NULL);
       HMAC (d->hash, skeyid_e, EVP_MD_size (d->hash), d->des_key,
	     EVP_MD_size (d->hash), d->des_key + EVP_MD_size (d->hash), NULL);
       memset (skeyid_e, 0, EVP_MD_size (d->hash));
     }

     /* Determine the initial 3DES IV.  */
     {
       EVP_MD_CTX ctx;
       unsigned char des_iv[EVP_MAX_MD_SIZE];
       EVP_DigestInit (&ctx, d->hash);
       EVP_DigestUpdate (&ctx, dh_public, sizeof (dh_public));
       EVP_DigestUpdate (&ctx, ke->u.ke.data, ke->u.ke.length);
       EVP_DigestFinal (&ctx, des_iv, NULL);
       memcpy (d->current_iv, des_iv, EVP_CIPHER_iv_length (d->cipher));
       memset (d->current_iv_msgid, 0, 4);
       memset (des_iv, 0, sizeof (des_iv));
     }
  }

  /* Send final phase 1 packet.  */
  {
    struct isakmp_packet *p2;
    uint8_t *p2kt;
    size_t p2kt_len;
    
    p2 = new_isakmp_packet ();
    memcpy (p2->i_cookie, d->i_cookie, ISAKMP_COOKIE_LENGTH);
    memcpy (p2->r_cookie, d->r_cookie, ISAKMP_COOKIE_LENGTH);
    p2->flags = ISAKMP_FLAG_E;
    p2->isakmp_version = ISAKMP_VERSION;
    p2->exchange_type = ISAKMP_EXCHANGE_AGGRESSIVE;
    p2->payload = new_isakmp_data_payload (ISAKMP_PAYLOAD_HASH,
					   returned_hash, 
					   EVP_MD_size (d->hash));
    flatten_isakmp_packet (p2, &p2kt, &p2kt_len);
    free_isakmp_packet (p2);
    isakmp_crypt (d, p2kt, p2kt_len, 1);

    memcpy (d->initial_iv, d->current_iv, EVP_CIPHER_iv_length (d->cipher));
    
    /* Now, send that packet and receive a new one.  */
    r_length = sendrecv (r_packet, sizeof (r_packet), 
			 p2kt, p2kt_len);
    free (p2kt);
  }
}

static uint16_t
unpack_verify_phase2 (struct sa_block *s, 
		      uint8_t *r_packet, 
		      size_t r_length,
		      struct isakmp_packet **r_p,
		      const uint8_t *nonce, size_t nonce_size)
{
  struct isakmp_packet *r;
  
  *r_p = NULL;
  
  if (r_length < ISAKMP_PAYLOAD_O 
      || ((r_length - ISAKMP_PAYLOAD_O) % EVP_CIPHER_block_size (s->cipher) 
	  != 0))
    return ISAKMP_N_UNEQUAL_PAYLOAD_LENGTHS;

  isakmp_crypt (s, r_packet, r_length, 0);
      
  {
    uint16_t reject = 0;
    r = parse_isakmp_packet (r_packet, r_length, &reject);
    if (reject != 0)
      return reject;
  }
	
  /* Verify the basic stuff.  */
  if (memcmp (r->i_cookie, s->i_cookie, ISAKMP_COOKIE_LENGTH) != 0
      || memcmp (r->r_cookie, s->r_cookie, ISAKMP_COOKIE_LENGTH) != 0)
    return ISAKMP_N_INVALID_COOKIE;
  if (r->flags != ISAKMP_FLAG_E)
    return ISAKMP_N_INVALID_FLAGS;
  
  {
    size_t sz, spos;
    HMAC_CTX ctx;
    unsigned char expected_hash[EVP_MAX_MD_SIZE];
    struct isakmp_payload *h = r->payload;
    
    if (h == NULL
	|| h->type != ISAKMP_PAYLOAD_HASH
	|| h->u.hash.length != EVP_MD_size(s->hash))
      return ISAKMP_N_INVALID_HASH_INFORMATION;
    
    spos = (ISAKMP_PAYLOAD_O
	    + (r_packet[ISAKMP_PAYLOAD_O + 2] << 8) 
	    + r_packet[ISAKMP_PAYLOAD_O + 3]);
    
    /* Compute the real length based on the payload lengths.  */
    for (sz = spos; 
	 r_packet[sz] != 0; 
	 sz += r_packet [sz+2] << 8 | r_packet[sz+3])
      ;
    sz += r_packet [sz+2] << 8 | r_packet[sz+3];
    
    HMAC_Init (&ctx, s->skeyid_a, EVP_MD_size(s->hash), s->hash);
    HMAC_Update (&ctx, r_packet + ISAKMP_MESSAGE_ID_O, 4);
    if (nonce)
      HMAC_Update (&ctx, nonce, nonce_size);
    HMAC_Update (&ctx, r_packet + spos, sz - spos);
    HMAC_Final (&ctx, expected_hash, NULL);
    
    if (memcmp (h->u.hash.data, expected_hash, EVP_MD_size (s->hash)) != 0)
      return ISAKMP_N_AUTHENTICATION_FAILED;
  }
  *r_p = r;
  return 0;
}

static void
phase2_authpacket (struct sa_block *s, struct isakmp_payload *pl,
		   uint8_t exchange_type, uint32_t msgid,
		   uint8_t **p_flat, size_t *p_size)
{
  struct isakmp_packet *p;
  uint8_t *pl_flat;
  size_t pl_size;
  HMAC_CTX ctx;
  uint8_t msgid_sent[4];

  /* Build up the packet.  */
  p = new_isakmp_packet();
  memcpy (p->i_cookie, s->i_cookie, ISAKMP_COOKIE_LENGTH);
  memcpy (p->r_cookie, s->r_cookie, ISAKMP_COOKIE_LENGTH);
  p->flags = ISAKMP_FLAG_E;
  p->isakmp_version = ISAKMP_VERSION;
  p->exchange_type = exchange_type;
  p->message_id = msgid;
  p->payload = new_isakmp_payload (ISAKMP_PAYLOAD_HASH);
  p->payload->next = pl;
  p->payload->u.hash.length = EVP_MD_size (s->hash);
  p->payload->u.hash.data = xallocc (EVP_MD_size (s->hash));
  
  /* Set the MAC.  */
  HMAC_Init (&ctx, s->skeyid_a, EVP_MD_size (s->hash), s->hash);

  msgid_sent[0] = msgid >> 24;
  msgid_sent[1] = msgid >> 16;
  msgid_sent[2] = msgid >> 8;
  msgid_sent[3] = msgid;
  HMAC_Update (&ctx, msgid_sent, sizeof (msgid_sent));

  flatten_isakmp_payload (pl, &pl_flat, &pl_size);
  HMAC_Update (&ctx, pl_flat, pl_size);
  memset (pl_flat, 0, pl_size);
  free (pl_flat);

  HMAC_Final (&ctx, p->payload->u.hash.data, NULL);

  flatten_isakmp_packet (p, p_flat, p_size);
  free_isakmp_packet (p);
}

static void
sendrecv_phase2 (struct sa_block *s, struct isakmp_payload *pl,
		 uint8_t exchange_type, uint32_t msgid, int sendonly)
{
  uint8_t *p_flat;
  size_t p_size;

  phase2_authpacket (s, pl, exchange_type, msgid, &p_flat, &p_size);
  isakmp_crypt (s, p_flat, p_size, 1);

  if (! sendonly)
    r_length = sendrecv (r_packet, sizeof (r_packet), 
			 p_flat, p_size);
  else
    {
      if (sendto (sockfd, p_flat, p_size, 0,
		  dest_addr, sizeof (struct sockaddr_in)) != (int) p_size
	  && sendonly == 1)
	error (1, errno, "can't send packet");
    }
  free (p_flat);
}

static void
phase2_fatal (struct sa_block *s, const char *msg, uint16_t id)
{
  struct isakmp_payload *pl;
  uint32_t msgid;

  get_random (&msgid, sizeof (msgid));
  pl = new_isakmp_payload (ISAKMP_PAYLOAD_N);
  pl->u.n.doi = ISAKMP_DOI_IPSEC;
  pl->u.n.protocol = ISAKMP_IPSEC_PROTO_ISAKMP;
  pl->u.n.type = id;
  sendrecv_phase2 (s, pl, ISAKMP_EXCHANGE_INFORMATIONAL, msgid, 2);
  
  get_random (&msgid, sizeof (msgid));
  pl = new_isakmp_payload (ISAKMP_PAYLOAD_D);
  pl->u.d.doi = ISAKMP_DOI_IPSEC;
  pl->u.d.protocol = ISAKMP_IPSEC_PROTO_ISAKMP;
  pl->u.d.num_spi = 1;
  sendrecv_phase2 (s, pl, ISAKMP_EXCHANGE_INFORMATIONAL, msgid, 2);

  error (1, 0, "%s: %s", msg, isakmp_notify_to_error (id));
}

static void 
do_phase_2_xauth (struct sa_block *s)
{
  struct isakmp_packet *r;
  int loopcount;

  /* This can go around for a while.  */
  for (loopcount = 0;; loopcount++)
    {
      uint16_t reject;
      struct isakmp_payload *rp;
      struct isakmp_attribute *a, *ap, *reply_attr;
      char ntop_buf[32];
      
      reject = unpack_verify_phase2 (s, r_packet, r_length, &r, NULL, 0);
      
      /* Check the transaction type is OK.  */
      if (reject == 0 && 
	  r->exchange_type != ISAKMP_EXCHANGE_MODECFG_TRANSACTION)
	reject = ISAKMP_N_INVALID_EXCHANGE_TYPE;
      
      /* After the hash, expect an attribute block.  */
      if (reject == 0
	  && (r->payload->next == NULL
	      || r->payload->next->next != NULL
	      || r->payload->next->type != ISAKMP_PAYLOAD_MODECFG_ATTR))
	reject = ISAKMP_N_INVALID_PAYLOAD_TYPE;
      
      if (reject == 0 &&
	  r->payload->next->u.modecfg.type == ISAKMP_MODECFG_CFG_SET)
	break;
      if (reject == 0
	  && r->payload->next->u.modecfg.type != ISAKMP_MODECFG_CFG_REQUEST)
	reject = ISAKMP_N_INVALID_PAYLOAD_TYPE;
      
      if (reject != 0)
	phase2_fatal (s, "expected xauth packet; rejected", reject);
      
      a = r->payload->next->u.modecfg.attributes;
      /* First, print any messages, and verify that we understand the
	 conversation.  */
      for (ap = a; ap && reject == 0; ap = ap->next)
	switch (ap->type)
	  {
	  case ISAKMP_XAUTH_ATTRIB_TYPE:
	    if (ap->af != isakmp_attr_16 || ap->u.attr_16 != 0)
	      reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
	    break;
	  case ISAKMP_XAUTH_ATTRIB_USER_NAME:
	  case ISAKMP_XAUTH_ATTRIB_USER_PASSWORD:
	  case ISAKMP_XAUTH_ATTRIB_PASSCODE:
	  case ISAKMP_XAUTH_ATTRIB_ANSWER:
	    break;
	  case ISAKMP_XAUTH_ATTRIB_MESSAGE:
	    if (ap->af == isakmp_attr_16)
	      printf ("%c%c\n", ap->u.attr_16 >> 8, ap->u.attr_16);
	    else
	      printf ("%.*s%s", ap->u.lots.length, ap->u.lots.data,
		      ((ap->u.lots.data
			&& ap->u.lots.data[ap->u.lots.length - 1] != '\n')
		       ? "\n" : ""));
	    break;
	  default:
	    reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
	  }
      if (reject != 0)
	phase2_fatal (s, "xauth packet unsupported", reject);
      
      inet_ntop (dest_addr->sa_family, 
		 &((struct sockaddr_in *)dest_addr)->sin_addr,
		 ntop_buf, sizeof (ntop_buf));
      
      /* Collect data from the user.  */
      reply_attr = NULL;
      for (ap = a; ap && reject == 0; ap = ap->next)
	switch (ap->type)
	  {
	  case ISAKMP_XAUTH_ATTRIB_USER_NAME:
	    if (loopcount == 0)
	      {
		struct isakmp_attribute *na;
		na = new_isakmp_attribute(ap->type, reply_attr);
		reply_attr = na;
		na->u.lots.length = strlen (config[CONFIG_XAUTH_USERNAME]) + 1;
		na->u.lots.data = xallocc (na->u.lots.length);
		strcpy (na->u.lots.data, config[CONFIG_XAUTH_USERNAME]);
		break;
	      }
	    printf ("User name for %s: ", ntop_buf);
	    fflush (stdout);
	  case ISAKMP_XAUTH_ATTRIB_ANSWER:
	    {
	      char *line = NULL;
	      size_t linelen = 0;
	      ssize_t linesz;
	      struct isakmp_attribute *na;
	      
	      if ((linesz = getline (&line, &linelen, stdin)) == -1)
		error (1, errno, "reading user input");
	      if (line[linesz - 1] == '\n')
		linesz--;
	      
	      na = new_isakmp_attribute(ap->type, reply_attr);
	      reply_attr = na;
	      na->u.lots.length = linesz;
	      na->u.lots.data = line;
	    }
	    break;
	    
	  case ISAKMP_XAUTH_ATTRIB_USER_PASSWORD:
	    if (loopcount == 0)
	      {
		struct isakmp_attribute *na;
		na = new_isakmp_attribute(ap->type, reply_attr);
		reply_attr = na;
		na->u.lots.length = strlen (config[CONFIG_XAUTH_PASSWORD]) + 1;
		na->u.lots.data = xallocc (na->u.lots.length);
		strcpy (na->u.lots.data, config[CONFIG_XAUTH_PASSWORD]);
		break;
	      }
	  case ISAKMP_XAUTH_ATTRIB_PASSCODE:
	    {
	      char prompt[64];
	      char *pass;
	      struct isakmp_attribute *na;
	      
	      if (ap->type == ISAKMP_XAUTH_ATTRIB_USER_PASSWORD)
		sprintf (prompt, "Password for VPN at %s: ", ntop_buf);
	      else
		sprintf (prompt, "Passcode for VPN at %s: ", ntop_buf);
	      pass = getpass (prompt);
	      
	      na = new_isakmp_attribute(ap->type, reply_attr);
	      reply_attr = na;
	      na->u.lots.length = strlen (pass) + 1;
	      na->u.lots.data = xallocc (na->u.lots.length);
	      memcpy (na->u.lots.data, pass, na->u.lots.length);
	      memset (pass, 0, na->u.lots.length);
	    }
	    break;
	  default:
	    ;
	  }
      
      /* Send the response.  */
      rp = new_isakmp_payload (ISAKMP_PAYLOAD_MODECFG_ATTR);
      rp->u.modecfg.type = ISAKMP_MODECFG_CFG_REPLY;
      rp->u.modecfg.id = r->payload->next->u.modecfg.id;
      rp->u.modecfg.attributes = reply_attr;
      sendrecv_phase2 (s, rp, ISAKMP_EXCHANGE_MODECFG_TRANSACTION,
		       r->message_id, 0);
      
      free_isakmp_packet (r);
    }
  
  {
    /* The final SET should have just one attribute.  */
    uint16_t reject = 0;
    struct isakmp_attribute *a = r->payload->next->u.modecfg.attributes;
    uint16_t set_result;
    
    if (a == NULL
	|| a->type != ISAKMP_XAUTH_ATTRIB_STATUS
	|| a->af != isakmp_attr_16
	|| a->next != NULL)
      {
	reject = ISAKMP_N_INVALID_PAYLOAD_TYPE;
	phase2_fatal (s, "xauth SET response rejected", reject);
      }
    set_result = a->u.attr_16;

    /* ACK the SET.  */
    r->payload->next->u.modecfg.type = ISAKMP_MODECFG_CFG_ACK;
    sendrecv_phase2 (s, r->payload->next, ISAKMP_EXCHANGE_MODECFG_TRANSACTION,
		     r->message_id, 1);
    r->payload->next = NULL;
    free_isakmp_packet (r);

    if (set_result == 0)
      error (2, 0, "authentication unsuccessful");
  }
}

static void 
do_phase_2_config (struct sa_block *s)
{
  struct isakmp_payload *rp;
  struct isakmp_attribute *a;
  struct isakmp_packet *r;
  uint32_t msgid;
  uint16_t reject;
  int seen_address = 0, seen_netmask = 0;
  
  get_random (&msgid, sizeof (msgid));
  if (msgid == 0)
    msgid = 1;
  
  rp = new_isakmp_payload (ISAKMP_PAYLOAD_MODECFG_ATTR);
  rp->u.modecfg.type = ISAKMP_MODECFG_CFG_REQUEST;
  rp->u.modecfg.id = 20;
  a = NULL;
  a = new_isakmp_attribute(ISAKMP_MODECFG_ATTRIB_INTERNAL_IP4_NETMASK, a);
  a = new_isakmp_attribute(ISAKMP_MODECFG_ATTRIB_INTERNAL_IP4_ADDRESS, a);
  rp->u.modecfg.attributes = a;
  sendrecv_phase2 (s, rp, ISAKMP_EXCHANGE_MODECFG_TRANSACTION,
		   msgid, 0);

  reject = unpack_verify_phase2 (s, r_packet, r_length, &r, NULL, 0);

  /* Check the transaction type & message ID are OK.  */
  if (reject == 0 && r->message_id != msgid)
    reject = ISAKMP_N_INVALID_MESSAGE_ID;
  if (reject == 0 && r->exchange_type != ISAKMP_EXCHANGE_MODECFG_TRANSACTION)
    reject = ISAKMP_N_INVALID_EXCHANGE_TYPE;
  
  /* After the hash, expect an attribute block.  */
  if (reject == 0
      && (r->payload->next == NULL
	  || r->payload->next->next != NULL
	  || r->payload->next->type != ISAKMP_PAYLOAD_MODECFG_ATTR
#if 0
	  || r->payload->next->u.modecfg.id != 20
#endif
	  || r->payload->next->u.modecfg.type != ISAKMP_MODECFG_CFG_REPLY))
    reject = ISAKMP_N_PAYLOAD_MALFORMED;

  if (reject != 0)
    phase2_fatal (s, "configuration response rejected", reject);

  for (a = r->payload->next->u.modecfg.attributes; 
       a && reject == 0; 
       a = a->next)
    switch (a->type)
      {
      case ISAKMP_MODECFG_ATTRIB_INTERNAL_IP4_NETMASK:
	if (a->af != isakmp_attr_lots || a->u.lots.length != 4)
	  reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
	else
	  memcpy (s->our_netmask, a->u.lots.data, 4);
	seen_netmask = 1;
	break;
	
      case ISAKMP_MODECFG_ATTRIB_INTERNAL_IP4_ADDRESS:
	if (a->af != isakmp_attr_lots || a->u.lots.length != 4)
	  reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
	else
	  memcpy (s->our_address, a->u.lots.data, 4);
	seen_address = 1;
	break;

      default:
	reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
	break;
      }
  
  if (reject == 0 && ! seen_address)
    reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;

  if (reject != 0)
    phase2_fatal (s, "configuration response rejected", reject);

  if (! seen_netmask)
    {
      uint32_t addr = (s->our_address[0] << 24 | s->our_address[1] << 16
		       | s->our_address[2] << 8 | s->our_address[3]);
      uint32_t mask = IN_CLASSC_NET;
      
      if (IN_CLASSA (addr))
	mask = IN_CLASSA_NET;
      if (IN_CLASSB (addr))
	mask = IN_CLASSB_NET;

      s->our_netmask[0] = mask >> 24;
      s->our_netmask[1] = mask >> 16;
      s->our_netmask[2] = mask >> 8;
      s->our_netmask[3] = mask;
    }
}

static void
config_tunnel (struct sa_block *s)
{
  int sock;
  struct ifreq ifr;
  uint8_t *addr;
  int i;
  
  sock = socket (AF_INET, SOCK_DGRAM, 0);
  if (sock < 0)
    error (1, errno, "making socket");
  memset (&ifr, 0, sizeof(ifr));
  memcpy (ifr.ifr_name, tun_name, IFNAMSIZ);
  ifr.ifr_addr.sa_family = AF_INET;
  addr = (uint8_t *)&((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
  memcpy (addr, s->our_address, 4);
  if (ioctl (sock, SIOCSIFADDR, &ifr) != 0)
    error (1, errno, "setting interface address");
  memcpy (addr, s->our_netmask, 4);
  if (ioctl (sock, SIOCSIFNETMASK, &ifr) != 0)
    error (1, errno, "setting interface netmask");
  for (i = 0; i < 4; i++)
    addr[i] = s->our_address[i] | ~s->our_netmask[i];
  if (ioctl (sock, SIOCSIFBRDADDR, &ifr) != 0)
    error (1, errno, "setting interface broadcast");
  /* The magic constants are:
     1500  the normal ethernet MTU
     -20   the size of an IP header
     -8    the size of an ESP header
     -2    minimum padding length
     -12   the size of the HMAC
  */
  ifr.ifr_mtu = 1500-20-8-2-12;
  if (ioctl (sock, SIOCSIFMTU, &ifr) != 0)
    error (1, errno, "setting interface MTU");
  if (ioctl (sock, SIOCGIFFLAGS, &ifr) != 0)
    error (1, errno, "getting interface flags");
  ifr.ifr_flags |= (IFF_UP | IFF_BROADCAST | IFF_RUNNING | IFF_MULTICAST);
  if (ioctl (sock, SIOCSIFFLAGS, &ifr) != 0)
    error (1, errno, "setting interface flags");
  close (sock);
}

static uint8_t *
gen_keymat (struct sa_block *s,
	    uint8_t protocol, uint32_t spi, 
	    const uint8_t *ni_data, size_t ni_size,
	    const uint8_t *nr_data, size_t nr_size)
{
  HMAC_CTX ctx;
  uint8_t *block;
  int i;
  int hsize = EVP_MD_size (s->hash);
  int blksz;
  int cnt;
  
  blksz = hsize + EVP_CIPHER_key_length (s->cipher);
  cnt = (blksz + hsize - 1) / hsize;
  block = xallocc (cnt * hsize);
  if (cnt < 1)
    abort ();

  for (i = 0; i < cnt; i++)
    {
      HMAC_Init (&ctx, s->skeyid_d, hsize, s->hash);
      if (i != 0)
	HMAC_Update (&ctx, block + (i-1) * hsize, hsize);
      HMAC_Update (&ctx, &protocol, 1);
      HMAC_Update (&ctx, (uint8_t *)&spi, sizeof (spi));
      HMAC_Update (&ctx, ni_data, ni_size);
      HMAC_Update (&ctx, nr_data, nr_size);
      HMAC_Final (&ctx, block + i * hsize, NULL);
    }
  return block;
}

static void
setup_link (struct sa_block *s)
{
  struct isakmp_payload *rp, *us, *them, *nonce_r = NULL;
  struct isakmp_attribute *a;
  struct isakmp_packet *r;
  uint32_t msgid;
  uint16_t reject;
  uint8_t nonce[20];
  
  get_random (&s->tous_esp_spi, sizeof (s->tous_esp_spi));
  rp = new_isakmp_payload (ISAKMP_PAYLOAD_SA);
  rp->u.sa.doi = ISAKMP_DOI_IPSEC;
  rp->u.sa.situation = ISAKMP_IPSEC_SIT_IDENTITY_ONLY;
  rp->u.sa.proposals = new_isakmp_payload (ISAKMP_PAYLOAD_P);
  rp->u.sa.proposals->u.p.spi_size = 4;
  rp->u.sa.proposals->u.p.spi = xallocc (4);
  /* The sadb_sa_spi field is already in network order.  */
  memcpy (rp->u.sa.proposals->u.p.spi, &s->tous_esp_spi, 4);
  rp->u.sa.proposals->u.p.prot_id = ISAKMP_IPSEC_PROTO_IPSEC_ESP;
  rp->u.sa.proposals->u.p.transforms = new_isakmp_payload (ISAKMP_PAYLOAD_T);
  rp->u.sa.proposals->u.p.transforms->u.t.id = ISAKMP_IPSEC_ESP_3DES;
  a = new_isakmp_attribute_16 (ISAKMP_IPSEC_ATTRIB_AUTH_ALG, 
			       IPSEC_AUTH_HMAC_MD5,
			       NULL);
  a = new_isakmp_attribute_16 (ISAKMP_IPSEC_ATTRIB_ENCAP_MODE,
			       IPSEC_ENCAP_TUNNEL, a);
  rp->u.sa.proposals->u.p.transforms->u.t.attributes = a;

  get_random (nonce, sizeof (nonce));
  rp->next = new_isakmp_data_payload (ISAKMP_PAYLOAD_NONCE,
				      nonce, sizeof (nonce));
  
  us = new_isakmp_payload (ISAKMP_PAYLOAD_ID);
  us->u.id.type = ISAKMP_IPSEC_ID_IPV4_ADDR;
  us->u.id.length = 4;
  us->u.id.data = xallocc(4);
  memcpy (us->u.id.data, s->our_address, sizeof (struct in_addr));
  them = new_isakmp_payload (ISAKMP_PAYLOAD_ID);
  them->u.id.type = ISAKMP_IPSEC_ID_IPV4_ADDR_SUBNET;
  them->u.id.length = 8;
  them->u.id.data = xallocc(8);
  rp->next->next = us;
  us->next = them;

  get_random (&msgid, sizeof (msgid));
  if (msgid == 0)
    msgid = 1;
  
  sendrecv_phase2 (s, rp, ISAKMP_EXCHANGE_IKE_QUICK,
		   msgid, 0);

  reject = unpack_verify_phase2 (s, r_packet, r_length, &r, 
				 nonce, sizeof (nonce));

  /* Check the transaction type & message ID are OK.  */
  if (reject == 0 && r->message_id != msgid)
    reject = ISAKMP_N_INVALID_MESSAGE_ID;
  if (reject == 0 && r->exchange_type != ISAKMP_EXCHANGE_IKE_QUICK)
    reject = ISAKMP_N_INVALID_EXCHANGE_TYPE;
  
  /* The SA payload must be second.  */
  if (reject == 0 && r->payload->next->type != ISAKMP_PAYLOAD_SA)
    reject = ISAKMP_N_INVALID_PAYLOAD_TYPE;

  if (reject != 0)
    phase2_fatal (s, "quick mode response rejected", reject);

  for (rp = r->payload->next; rp && reject == 0; rp = rp->next)
    switch (rp->type)
      {
      case ISAKMP_PAYLOAD_SA:
	if (reject == 0 && rp->u.sa.doi != ISAKMP_DOI_IPSEC)
	  reject = ISAKMP_N_DOI_NOT_SUPPORTED;
	if (reject == 0 && 
	    rp->u.sa.situation != ISAKMP_IPSEC_SIT_IDENTITY_ONLY)
	  reject = ISAKMP_N_SITUATION_NOT_SUPPORTED;
	if (reject == 0 && 
	    (rp->u.sa.proposals == NULL
	     || rp->u.sa.proposals->next != NULL))
	  reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
	if (reject == 0 &&
	    rp->u.sa.proposals->u.p.prot_id != ISAKMP_IPSEC_PROTO_IPSEC_ESP)
	  reject = ISAKMP_N_INVALID_PROTOCOL_ID;
	if (reject == 0 &&
	    rp->u.sa.proposals->u.p.spi_size != 4)
	  reject = ISAKMP_N_INVALID_SPI;
	if (reject == 0 &&
	    (rp->u.sa.proposals->u.p.transforms == NULL
	     || rp->u.sa.proposals->u.p.transforms->next != NULL))
	  reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
	if (reject == 0 &&
	    (rp->u.sa.proposals->u.p.transforms->u.t.id 
	     != ISAKMP_IPSEC_ESP_3DES))
	  reject = ISAKMP_N_INVALID_TRANSFORM_ID;
	if (reject == 0) {
	  struct isakmp_attribute *a 
	    = rp->u.sa.proposals->u.p.transforms->u.t.attributes;
	  int seen_auth = 0, seen_encap = 0;

	  memcpy (&s->tothem_esp_spi, rp->u.sa.proposals->u.p.spi, 4);

	  for (; a && reject == 0; a = a->next)
	    switch (a->type)
	      {
	      case ISAKMP_IPSEC_ATTRIB_AUTH_ALG:
		if (a->af == isakmp_attr_16 &&
		    a->u.attr_16 == IPSEC_AUTH_HMAC_MD5)
		  seen_auth = 1;
		else
		  reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
		break;
	      case ISAKMP_IPSEC_ATTRIB_ENCAP_MODE:
		if (a->af == isakmp_attr_16 &&
		    a->u.attr_16 == IPSEC_ENCAP_TUNNEL)
		  seen_encap = 1;
		else
		  reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
		break;
	      default:
		reject = ISAKMP_N_ATTRIBUTES_NOT_SUPPORTED;
		break;
	      }
	  if (reject == 0 && (! seen_auth || ! seen_encap))
	    reject = ISAKMP_N_BAD_PROPOSAL_SYNTAX;
	}
	break;
	
      case ISAKMP_PAYLOAD_ID:		
	break;
      case ISAKMP_PAYLOAD_NONCE:	
	nonce_r = rp;
	break;

      default:
	reject = ISAKMP_N_INVALID_PAYLOAD_TYPE;
	break;
      }
  
  if (reject == 0 && nonce_r == NULL)
    reject = ISAKMP_N_INVALID_HASH_INFORMATION;
  if (reject != 0)
    phase2_fatal (s, "quick mode response rejected [2]", reject);
  
  /* Create the delete payload, now that we have all the information.  */
  {
    struct isakmp_payload *d_isakmp, *d_ipsec;
    uint32_t del_msgid;

    get_random (&del_msgid, sizeof (del_msgid));
    d_isakmp = new_isakmp_payload (ISAKMP_PAYLOAD_D);
    d_isakmp->u.d.doi = ISAKMP_DOI_IPSEC;
    d_isakmp->u.d.protocol = ISAKMP_IPSEC_PROTO_ISAKMP;
    d_isakmp->u.d.spi_length = 0;
    d_isakmp->u.d.num_spi = 1;
    d_ipsec = new_isakmp_payload (ISAKMP_PAYLOAD_D);
    d_ipsec->next = d_isakmp;
    d_ipsec->u.d.doi = ISAKMP_DOI_IPSEC;
    d_ipsec->u.d.protocol = ISAKMP_IPSEC_PROTO_IPSEC_ESP;
    d_ipsec->u.d.spi_length = 4;
    d_ipsec->u.d.num_spi = 2;
    d_ipsec->u.d.spi = xallocc (2 * sizeof (uint8_t *));
    d_ipsec->u.d.spi[0] = xallocc (d_ipsec->u.d.spi_length);
    memcpy (d_ipsec->u.d.spi[0], &s->tous_esp_spi, 4);
    d_ipsec->u.d.spi[1] = xallocc (d_ipsec->u.d.spi_length);
    memcpy (d_ipsec->u.d.spi[1], &s->tothem_esp_spi, 4);
    phase2_authpacket (s, d_ipsec, ISAKMP_EXCHANGE_INFORMATIONAL,
		       del_msgid, &s->kill_packet, &s->kill_packet_size);
    isakmp_crypt (s, s->kill_packet, s->kill_packet_size, -1);
  }
  
  /* Set up the interface here so it's ready when our acknowledgement
     arrives.  */
  {
    uint8_t *tous_keys, *tothem_keys;
    struct sockaddr_in tothem_dest;
    
    tous_keys = gen_keymat (s, ISAKMP_IPSEC_PROTO_IPSEC_ESP,
			    s->tous_esp_spi, nonce, sizeof (nonce), 
			    nonce_r->u.nonce.data, nonce_r->u.nonce.length);
    memset (&tothem_dest, 0, sizeof (tothem_dest));
    tothem_dest.sin_family = AF_INET;
    memcpy (&tothem_dest.sin_addr, s->our_address, 4);
    tothem_keys = gen_keymat (s, ISAKMP_IPSEC_PROTO_IPSEC_ESP,
			      s->tothem_esp_spi, nonce, sizeof (nonce), 
			      nonce_r->u.nonce.data, nonce_r->u.nonce.length);
    vpnc_doit (s->tous_esp_spi, tous_keys, &tothem_dest,
	       s->tothem_esp_spi, tothem_keys, (struct sockaddr_in *)dest_addr,
	       tun_fd, random_fd, sockfd,
	       s->kill_packet, s->kill_packet_size, dest_addr);
  }
  
  {
    struct isakmp_packet *p;
    uint8_t *p_flat;
    size_t p_size;
    HMAC_CTX ctx;
    uint8_t msgid_sent[4];
    static const uint8_t zero = 0;

    /* Build up the packet.  */
    p = new_isakmp_packet();
    memcpy (p->i_cookie, s->i_cookie, ISAKMP_COOKIE_LENGTH);
    memcpy (p->r_cookie, s->r_cookie, ISAKMP_COOKIE_LENGTH);
    p->flags = ISAKMP_FLAG_E;
    p->isakmp_version = ISAKMP_VERSION;
    p->exchange_type = ISAKMP_EXCHANGE_IKE_QUICK;
    p->message_id = msgid;
    p->payload = new_isakmp_payload (ISAKMP_PAYLOAD_HASH);
    p->payload->u.hash.length = EVP_MD_size (s->hash);
    p->payload->u.hash.data = xallocc (EVP_MD_size (s->hash));
  
    /* Set the MAC.  */
    HMAC_Init (&ctx, s->skeyid_a, EVP_MD_size (s->hash), s->hash);
    HMAC_Update (&ctx, &zero, 1);
    msgid_sent[0] = msgid >> 24;
    msgid_sent[1] = msgid >> 16;
    msgid_sent[2] = msgid >> 8;
    msgid_sent[3] = msgid;
    HMAC_Update (&ctx, msgid_sent, sizeof (msgid_sent));
    HMAC_Update (&ctx, nonce, sizeof (nonce));
    HMAC_Update (&ctx, nonce_r->u.nonce.data, nonce_r->u.nonce.length);
    HMAC_Final (&ctx, p->payload->u.hash.data, NULL);

    /* Send the packet.  FIXME: what happens if it gets lost?  */
    flatten_isakmp_packet (p, &p_flat, &p_size);
    free_isakmp_packet (p);
    isakmp_crypt (s, p_flat, p_size, 1);
    if (sendto (sockfd, p_flat, p_size, 0,
		dest_addr, sizeof (struct sockaddr_in)) != (int) p_size)
      error (1, errno, "can't send packet");
    free (p_flat);
  }
}

static const struct config_names_s {
  const char *name;
  const char *option;
  enum config_enum nm;
} config_names[] = {
  { "IPSec gateway ", "--gateway", CONFIG_IPSEC_GATEWAY },
  { "IPSec ID ", "--id", CONFIG_IPSEC_ID },
  { "IPSec secret ", NULL, CONFIG_IPSEC_SECRET },
  { "Xauth username ", "--username", CONFIG_XAUTH_USERNAME },
  { "Xauth password ", NULL, CONFIG_XAUTH_PASSWORD },
  { NULL, NULL, 0 }
};

void
read_config_file (char *name, const char **configs, int missingok)
{
  FILE *f;
  char *line = NULL;
  ssize_t line_length = 0;
  int linenum = 0;
  
  f = fopen (name, "r");
  if (missingok && f == NULL && errno == ENOENT)
    return;
  if (f == NULL)
    error (1, errno, "couldn't open `%s'", name);
  for (;;)
    {
      ssize_t llen;
      int i;
      
      llen = getline (&line, &line_length, f);
      if (llen == -1 && feof (f))
	break;
      if (llen == -1)
	error (1, errno, "reading `%s'", name);
      if (line[llen - 1] == '\n')
	line[llen - 1] = 0;
      linenum++;
      for (i = 0; config_names[i].name != NULL; i++)
	if (strncasecmp (config_names[i].name, line, 
			 strlen (config_names[i].name)) == 0)
	  {
	    if (configs[config_names[i].nm] == NULL)
	      configs[config_names[i].nm] = 
		strdup (line + strlen (config_names[i].name));
	    if (configs[config_names[i].nm] == NULL)
	      error (1, errno, "can't allocate memory");
	    break;
	  }
      if (config_names[i].name == NULL && line[0] != '#' && line[0] != 0)
	error_at_line (0, 0, name, linenum, 
		       "warning: unknown configuration directive");
    }
}

int main(int argc, char **argv)
{
  struct sa_block oursa;
  int i;
  int print_config = 0;
  
  test_pack_unpack();

  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
      {
	int c;
	int known = 0;

	for (c = 0; config_names[c].name != NULL && ! known; c++)
	  if (config_names[c].option != NULL
	      && strncmp (argv[i], config_names[c].option,
			  strlen (config_names[c].option)) == 0)
	    {
	      char *s;
	      
	      known = 1;
	      if (argv[i][strlen (config_names[c].option)] == '=')
		s = strdup (argv[i] + strlen (config_names[c].option) + 1);
	      else if (argv[i][strlen (config_names[c].option)] == 0
		       && i + 1 < argc)
		s = argv[++i];
	      else
		known = 0;
	      if (known)
		config[config_names[c].nm] = s;
	    }
	
	if (! known && strcmp (argv[i], "--version") == 0)
	  {
	    printf ("vpnc version 0.1\n");
	    printf ("Copyright (C) 2002 Geoffrey Keating\n");
	    printf ("%s",
"vpnc comes with NO WARRANTY, to the extent permitted by law.\n"
"You may redistribute copies of vpnc under the terms of the GNU General\n"
"Public License.  For more information about these matters, see the files\n"
		    "named COPYING.\n");
	    exit (0);
	  }
	if (! known && strcmp (argv[i], "--print-config") == 0)
	  {
	    print_config = 1;
	    known = 1;
	    break;
	  }
	
	if (! known)
	  {
	    int c;
	    
	    printf ("usage: %s [options] [config file]\n", argv[0]);
	    printf ("  %14s %s\n", "Option", "Argument");
	    for (c = 0; config_names[c].name != NULL; c++)
	      printf ("  %14s %s\n", 
		      (config_names[c].option == NULL
		       ? "(no option)"
		       : config_names[c].option),
		      config_names[c].name);
	    printf ("Report bugs to geoffk@geoffk.org\n");
	    exit (0);
	  }
      }
    else
      read_config_file (argv[i], config, 0);

  read_config_file ("/etc/vpnc.conf", config, 1);

  for (i = 0; i < LAST_CONFIG; i++)
    if (config[i] == NULL)
      {
	char *s = NULL;
	size_t s_len = 0;
	
	switch (i)
	  {
	  case CONFIG_IPSEC_GATEWAY:
	    printf ("Enter IPSec gateway address: ");
	    break;
	  case CONFIG_IPSEC_ID:
	    printf ("Enter IPSec ID for %s: ", 
		    config[CONFIG_IPSEC_GATEWAY]);
	    break;
	  case CONFIG_IPSEC_SECRET:
	    printf ("Enter IPSec secret for %s@%s: ",
		    config[CONFIG_IPSEC_ID], config[CONFIG_IPSEC_GATEWAY]);
	    break;
	  case CONFIG_XAUTH_USERNAME:
	    printf ("Enter username for %s: ", config[CONFIG_IPSEC_GATEWAY]);
	    break;
	  case CONFIG_XAUTH_PASSWORD:
	    printf ("Enter password for %s@%s: ",
		    config[CONFIG_XAUTH_USERNAME],
		    config[CONFIG_IPSEC_GATEWAY]);
	    break;
	  }
	fflush (stdout);
	if (i == CONFIG_IPSEC_SECRET || i == CONFIG_XAUTH_PASSWORD)
	  {
	    s = strdup (getpass (""));
	  }
	else
	  getline (&s, &s_len, stdin);
	if (s != NULL && s[strlen (s) - 1] == '\n')
	  s[strlen (s) - 1] = 0;
	config[i] = s;
      }
  
  if (print_config)
    {
      for (i = 0; config_names[i].name != NULL; i++)
	printf ("%s%s\n", config_names[i].name, config[config_names[i].nm]);
      exit (0);
    }

  dest_addr = init_sockaddr (config[CONFIG_IPSEC_GATEWAY], 500);
  sockfd = make_socket (500);
  setup_tunnel();

  do_phase_1 (config[CONFIG_IPSEC_ID], config[CONFIG_IPSEC_SECRET], &oursa);
  do_phase_2_xauth (&oursa);
  do_phase_2_config (&oursa);
  config_tunnel (&oursa);

  setup_link (&oursa);

  return 0;
}
